from nawah.config import Config
from nawah.enums import Event, NAWAH_VALUES, LOCALE_STRATEGY
from nawah.utils import _extract_attr
from nawah.classes import (
	ATTR,
	NAWAH_ENV,
	NAWAH_QUERY,
	Query,
	NAWAH_DOC,
	BaseModel,
	DictObj,
	ATTRS_TYPES_ARGS,
	NAWAH_EVENTS,
	InvalidAttrTypeException,
	MissingAttrException,
	InvalidAttrException,
	ConvertAttrException,
)

from bson import binary, ObjectId
from typing import Dict, Optional, List, Union, Any, cast, Literal, Tuple, TYPE_CHECKING

import logging, copy, re, asyncio, datetime

if TYPE_CHECKING:
	from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


async def _process_file_obj(
	*, doc: Union[NAWAH_DOC, dict, list], modules: Dict[str, 'BaseModule'], env: NAWAH_ENV
):
	if type(doc) == dict:
		doc_iter = doc.keys()  # type: ignore
	elif type(doc) == list:
		doc_iter = range(len(doc))
	for j in doc_iter:
		if type(doc[j]) == dict:  # type: ignore
			if '__file' in doc[j].keys():  # type: ignore
				file_id = doc[j]['__file']  # type: ignore
				logger.debug(
					f'Detected file in doc. Retrieving file from File module with _id: \'{file_id}\'.'
				)
				try:
					file_results = await modules['file'].read(
						skip_events=[Event.PERM], env=env, query=[{'_id': file_id}]
					)
					doc[j] = file_results.args.docs[0].file  # type: ignore
					file_results = await modules['file'].delete(
						skip_events=[Event.PERM, Event.SOFT],
						env=env,
						query=[{'_id': file_id}],
					)
					if file_results.status != 200 or file_results.args.count != 1:
						logger.warning(
							f'Filed to delete doc _id \'{file_id}\' from File module after retrieving.'
						)
				except Exception as e:
					logger.error(f'Failed to retrieve doc _id \'{file_id}\', with error:')
					logger.error(e)
					doc[j] = None  # type: ignore
			else:
				await _process_file_obj(doc=doc[j], modules=modules, env=env)  # type: ignore
		elif type(doc[j]) == list:  # type: ignore
			await _process_file_obj(doc=doc[j], modules=modules, env=env)  # type: ignore


async def validate_doc(
	*,
	mode: Literal['create', 'create_draft', 'update'],
	doc: NAWAH_DOC,
	attrs: Dict[str, ATTR],
	skip_events: NAWAH_EVENTS = None,
	env: NAWAH_ENV = None,
	query: Union[NAWAH_QUERY, Query] = None,
):
	attrs_map = {attr.split('.')[0]: attr for attr in doc.keys()}

	for attr in attrs.keys():
		if attr not in attrs_map.keys():
			if mode == 'create':
				doc[attr] = None
			else:
				continue
		elif mode != 'create':
			attr = attrs_map[attr]

		try:
			env = cast(NAWAH_ENV, env)
			if mode != 'create' and '.' in attr:
				doc[attr] = await _validate_dot_notated(
					attr=attr,
					doc=doc,
					attrs=attrs,
					skip_events=skip_events,
					env=env,
					query=query,
				)
			else:
				doc[attr] = await validate_attr(
					attr_name=attr,
					attr_type=attrs[attr],
					attr_val=doc[attr],
					mode=mode,
					skip_events=skip_events,
					env=env,
					query=query,
					doc=doc,
				)
		except Exception as e:
			if type(e) in [InvalidAttrException, ConvertAttrException]:
				if doc[attr] == None:
					raise MissingAttrException(attr_name=attr)
				else:
					raise e
			else:
				raise e


async def _validate_dot_notated(
	attr: str,
	doc: NAWAH_DOC,
	attrs: Dict[str, ATTR],
	skip_events: Optional[NAWAH_EVENTS],
	env: Optional[NAWAH_ENV],
	query: Optional[Union[NAWAH_QUERY, Query]],
):
	attr_path = attr.split('.')
	attr_path_len = len(attr_path)
	attr_type: Union[Dict[str, ATTR], ATTR] = attrs

	try:
		for i in range(attr_path_len):
			# [DOC] Iterate over attr_path to reach last valid Attr Type
			if type(attr_type) == dict:
				attr_type = cast(dict, attr_type)
				attr_type = attr_type[attr_path[i]]
			elif type(attr_type) == ATTR:
				attr_type = cast(ATTR, attr_type)
				if attr_type._type == 'ANY':
					return doc[attr]
				elif attr_type._type == 'LOCALE':
					if attr_path[i] not in Config.locales:
						raise Exception()
					attr_type = ATTR.STR()
				elif attr_type._type == 'TYPED_DICT':
					attr_type = attr_type._args['dict'][attr_path[i]]
				elif attr_type._type == 'KV_DICT':
					attr_type = attr_type._args['val']
				# [DOC] However, if list or union, start a new _validate_dot_notated call as it is required to check all the provided types
				elif attr_type._type in ['LIST', 'UNION']:
					if attr_type._type == 'LIST':
						attr_type_iter = attr_type._args['list']
					else:
						attr_type_iter = attr_type._args['union']
					for child_attr_type in attr_type_iter:
						attr_val = await _validate_dot_notated(
							attr='.'.join(attr_path[i:]),
							doc={'.'.join(attr_path[i:]): doc[attr]},
							attrs={attr_path[i]: child_attr_type},
							skip_events=skip_events,
							env=env,
							query=query,
						)
						if attr_val != None:
							return attr_val
					raise Exception()
				else:
					raise Exception()
			else:
				raise Exception()

		attr_type = cast(ATTR, attr_type)
		# [DOC] Validate val against final Attr Type
		# [DOC] mode is statically set to update as dot-notation attrs are only allowed in update calls
		env = cast(NAWAH_ENV, env)
		attr_val = await validate_attr(
			mode='update',
			attr_name=attr,
			attr_type=attr_type,
			attr_val=doc[attr],
			skip_events=skip_events,
			env=env,
			query=query,
			doc=doc,
		)
		return attr_val
	except:
		raise InvalidAttrException(
			attr_name=attr, attr_type=attrs[attr_path[0]], val_type=type(doc[attr])
		)


async def _validate_default(
	*,
	mode: Literal['create', 'create_draft', 'update', 'deep'],
	attr_name: str,
	attr_type: ATTR,
	attr_val: Any,
	skip_events: NAWAH_EVENTS,
	env: NAWAH_ENV,
	query: Union[NAWAH_QUERY, Query],
	doc: NAWAH_DOC,
	scope: NAWAH_DOC,
):
	if (
		mode == 'create'
		and type(attr_type._default) == ATTR
		and attr_type._default._type == 'TYPE'
	):
		attr_val = await attr_type._default._args['func'](
			mode=mode,
			attr_name=attr_name,
			attr_type=attr_type,
			attr_val=attr_val,
			skip_events=skip_events,
			env=env,
			query=query,
			doc=doc,
			scope=scope,
		)
		return copy.deepcopy(attr_val)

	elif attr_type._type == 'COUNTER':
		counter_groups = re.findall(
			r'(\$__(?:values:[0-9]+|counters\.[a-z0-9_]+))', attr_type._args['pattern']
		)
		counter_val = attr_type._args['pattern']
		for group in counter_groups:
			for group in counter_groups:
				if group.startswith('$__values:'):
					value_callable = attr_type._args['values'][int(group.replace('$__values:', ''))]
					counter_val = counter_val.replace(
						group, str(value_callable(skip_events=skip_events, env=env, query=query, doc=doc))
					)
				elif group.startswith('$__counters.'):
					env = cast(NAWAH_ENV, env)
					counter_name = group.replace('$__counters.', '')
					setting_read_results = await Config.modules['setting'].read(
						skip_events=[Event.PERM],
						env=env,
						query=[
							{
								'var': '__counter:' + counter_name,
								'type': 'global',
							}
						],
					)
					setting = setting_read_results.args.docs[0]
					setting_update_results = asyncio.create_task(
						Config.modules['setting'].update(
							skip_events=[Event.PERM],
							env=env,
							query=[{'_id': setting._id, 'type': 'global'}],
							doc={'val': {'$add': 1}},
						)
					)
					# [DOC] Condition "not task.cancelled()" is added to avoid exceptions with the task getting cancelled during its run as such it might be running in test mode, or at time of shutting down Nawah
					setting_update_results.add_done_callback(
						setting_update_callback_wrapper(counter_name)
					)
					counter_val = counter_val.replace(group, str(setting.val + 1))
		return counter_val

	elif attr_val == None:
		if mode == 'update':
			return NAWAH_VALUES.NONE_VALUE
		elif attr_type._default != NAWAH_VALUES.NONE_VALUE:
			return copy.deepcopy(attr_type._default)

	raise Exception('No default set to validate.')


def setting_update_callback_wrapper(counter_name):
	def setting_update_callback(task):
		if not task.cancelled() and task.result().status != 200:
			logger.error(f'Failed to update Setting doc for counter \'{counter_name}\'')

	return setting_update_callback


async def validate_attr(
	*,
	mode: Literal['create', 'create_draft', 'update', 'deep'],
	attr_name: str,
	attr_type: ATTR,
	attr_val: Any,
	skip_events: NAWAH_EVENTS = None,
	env: NAWAH_ENV = None,
	query: Union[NAWAH_QUERY, Query] = None,
	doc: NAWAH_DOC = None,
	scope: NAWAH_DOC = None,
):
	# [TODO] Test setting attr_name to reflect a path for nested attrs
	try:
		skip_events = cast(NAWAH_EVENTS, skip_events)
		env = cast(NAWAH_ENV, env)
		query = cast(Union[NAWAH_QUERY, Query], query)
		doc = cast(NAWAH_DOC, doc)
		return await _validate_default(
			mode=mode,
			attr_name=attr_name,
			attr_type=attr_type,
			attr_val=attr_val,
			skip_events=skip_events,
			env=env,
			query=query,
			doc=doc,
			scope=scope if scope else doc,
		)
	except:
		pass

	attr_oper: Literal[
		None, '$add', '$multiply', '$append', '$set_index', '$del_val', '$del_index'
	] = None
	attr_oper_args = {}
	if mode == 'update' and type(attr_val) == dict:
		if '$add' in attr_val.keys():
			attr_oper = '$add'
			if '$field' in attr_val.keys() and attr_val['$field']:
				attr_oper_args['$field'] = attr_val['$field']
			else:
				attr_oper_args['$field'] = None
			attr_val = attr_val['$add']
		elif '$multiply' in attr_val.keys():
			attr_oper = '$multiply'
			if '$field' in attr_val.keys() and attr_val['$field']:
				attr_oper_args['$field'] = attr_val['$field']
			else:
				attr_oper_args['$field'] = None
			attr_val = attr_val['$multiply']
		elif '$append' in attr_val.keys():
			attr_oper = '$append'
			if '$unique' in attr_val.keys() and attr_val['$unique'] == True:
				attr_oper_args['$unique'] = True
			else:
				attr_oper_args['$unique'] = False
			attr_val = [attr_val['$append']]
		elif '$set_index' in attr_val.keys():
			attr_oper = '$set_index'
			attr_oper_args['$index'] = attr_val['$index']
			attr_val = [attr_val['$set_index']]
		elif '$del_val' in attr_val.keys():
			attr_oper = '$del_val'
			attr_val = attr_val['$del_val']
			if attr_type._type != 'LIST' or type(attr_val) != list:
				raise InvalidAttrException(
					attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
				)
			return return_valid_attr(
				attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
			)
		elif '$del_index' in attr_val.keys():
			attr_oper = '$del_index'
			attr_oper_args['$index'] = attr_val['$del_index']
			attr_val = attr_val['$del_index']
			if (attr_type._type == 'LIST' and type(attr_val) == int) or (
				attr_type._type == 'KV_DICT' and type(attr_val) == str
			):
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)
			else:
				raise InvalidAttrException(
					attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
				)

	# [DOC] Deepcopy attr_val to eliminate changes in in original object
	attr_val = copy.deepcopy(attr_val)

	try:
		if attr_type._type == 'ANY':
			if attr_val != None:
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'ACCESS':
			if (
				type(attr_val) == dict
				and set(attr_val.keys()) == {'anon', 'users', 'groups'}
				and type(attr_val['anon']) == bool
				and type(attr_val['users']) == list
				and type(attr_val['groups']) == list
			):
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'BOOL':
			if type(attr_val) == bool:
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'DATE':
			if re.match(r'^[0-9]{4}-[0-9]{2}-[0-9]{2}$', attr_val):
				if attr_type._args['ranges']:
					for date_range in attr_type._args['ranges']:
						date_range = copy.deepcopy(date_range)
						for i in [0, 1]:
							if date_range[i][0] in ['+', '-']:
								date_range_delta = {}
								if date_range[i][-1] == 'd':
									date_range_delta = {'days': int(date_range[i][:-1])}
								elif date_range[i][-1] == 'w':
									date_range_delta = {'weeks': int(date_range[i][:-1])}
								date_range[i] = (
									(datetime.datetime.utcnow() + datetime.timedelta(**date_range_delta))
									.isoformat()
									.split('T')[0]
								)
						if attr_val >= date_range[0] and attr_val < date_range[1]:
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'DATETIME':
			if re.match(
				r'^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$',
				attr_val,
			):
				if attr_type._args['ranges']:
					for datetime_range in attr_type._args['ranges']:
						datetime_range = copy.deepcopy(datetime_range)
						for i in [0, 1]:
							if datetime_range[i][0] in ['+', '-']:
								datetime_range_delta = {}
								if datetime_range[i][-1] == 'd':
									datetime_range_delta = {'days': int(datetime_range[i][:-1])}
								elif datetime_range[i][-1] == 's':
									datetime_range_delta = {'seconds': int(datetime_range[i][:-1])}
								elif datetime_range[i][-1] == 'm':
									datetime_range_delta = {'minutes': int(datetime_range[i][:-1])}
								elif datetime_range[i][-1] == 'h':
									datetime_range_delta = {'hours': int(datetime_range[i][:-1])}
								elif datetime_range[i][-1] == 'w':
									datetime_range_delta = {'weeks': int(datetime_range[i][:-1])}
								datetime_range[i] = (
									datetime.datetime.utcnow() + datetime.timedelta(**datetime_range_delta)
								).isoformat()
						if attr_val >= datetime_range[0] and attr_val < datetime_range[1]:
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'DYNAMIC_ATTR':
			if type(attr_val) == dict:
				try:
					if (not attr_type._args['types']) or (
						attr_type._args['types'] and attr_val['type'] in attr_type._args['types']
					):
						_, attr_val = generate_dynamic_attr(dynamic_attr=attr_val)
						return return_valid_attr(
							attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
						)
				except Exception as e:
					logger.debug(
						'Exception occurred while validating type \'DYNAMIC_ATTR\'. Exception details:'
					)
					logger.debug(e)

		elif attr_type._type == 'DYNAMIC_VAL':
			# [DOC] Populate setting_query
			setting_query = {}
			if attr_type._args['dynamic_attr'].startswith('$__settings.global/'):
				setting_query['type'] = 'global'
				setting_query['var'] = attr_type._args['dynamic_attr'].split('/')[1]
			elif attr_type._args['dynamic_attr'].startswith('$__settings.user/'):
				setting_query['type'] = 'user'
				_, setting_query['user'], setting_query['var'] = attr_type._args[
					'dynamic_attr'
				].split('/')
			# [DOC] Check if variables are present in setting_query['var']
			for setting_query_var in re.findall(
				r'(\$__doc\.([a-zA-Z0-9_]+))', setting_query['var']
			):
				doc = cast(Dict[str, Any], doc)
				setting_query['var'] = setting_query['var'].replace(
					setting_query_var[0], str(_extract_attr(scope=doc, attr_path=setting_query_var[1]))
				)
			# [DOC] Read setting val
			env = cast(NAWAH_ENV, env)
			setting_results = await Config.modules['setting'].read(
				skip_events=[Event.PERM], env=env, query=[setting_query]
			)
			setting = setting_results.args.docs[0]
			dynamic_attr = generate_dynamic_attr(dynamic_attr=setting.val)[0]
			attr_val = await validate_attr(
				mode='create', attr_name=attr_name, attr_type=dynamic_attr, attr_val=attr_val
			)
			return return_valid_attr(
				attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
			)

		elif attr_type._type == 'KV_DICT':
			if type(attr_val) == dict:
				if attr_type._args['min']:
					if len(attr_val.keys()) < attr_type._args['min']:
						raise InvalidAttrException(
							attr_name=attr_name,
							attr_type=attr_type,
							val_type=type(attr_val),
						)
				if attr_type._args['max']:
					if len(attr_val.keys()) > attr_type._args['max']:
						raise InvalidAttrException(
							attr_name=attr_name,
							attr_type=attr_type,
							val_type=type(attr_val),
						)
				if attr_type._args['req']:
					for req_key in attr_type._args['req']:
						if req_key not in attr_val.keys():
							raise InvalidAttrException(
								attr_name=attr_name,
								attr_type=attr_type,
								val_type=type(attr_val),
							)
				shadow_attr_val = {}
				for child_attr_val in attr_val.keys():
					shadow_attr_val[
						await validate_attr(
							mode='deep',
							attr_name=f'{attr_name}.{child_attr_val}',
							attr_type=attr_type._args['key'],
							attr_val=child_attr_val,
							skip_events=skip_events,
							env=env,
							query=query,
							doc=doc,
							scope=attr_val,
						)
					] = await validate_attr(
						mode='deep',
						attr_name=f'{attr_name}.{child_attr_val}',
						attr_type=attr_type._args['val'],
						attr_val=attr_val[child_attr_val],
						skip_events=skip_events,
						env=env,
						query=query,
						doc=doc,
						scope=attr_val,
					)
				return return_valid_attr(
					attr_val=shadow_attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'TYPED_DICT':
			if type(attr_val) == dict:
				for child_attr_type in attr_type._args['dict'].keys():
					if child_attr_type not in attr_val.keys():
						attr_val[child_attr_type] = None
					try:
						attr_val[child_attr_type] = await validate_attr(
							mode='deep',
							attr_name=f'{attr_name}.{child_attr_type}',
							attr_type=attr_type._args['dict'][child_attr_type],
							attr_val=attr_val[child_attr_type],
							skip_events=skip_events,
							env=env,
							query=query,
							doc=doc,
							scope=attr_val,
						)
					except Exception as e:
						logger.debug(
							'Exception occurred while validating type \'TYPED_DICT\'. Exception details:'
						)
						logger.debug(e)
						raise e

				# [DOC] Match keys _after_ checking child attrs in order to allow _validate_default to run on all child attrs
				if set(attr_val.keys()) != set(attr_type._args['dict'].keys()):
					raise InvalidAttrException(
						attr_name=attr_name,
						attr_type=attr_type,
						val_type=type(attr_val),
					)

				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'EMAIL':
			if type(attr_val) == str and re.match(r'^[^@]+@[^@]+\.[^@]+$', attr_val):
				if attr_type._args['allowed_domains']:
					for domain in attr_type._args['allowed_domains']:
						if attr_type._args['strict']:
							domain = f'@{domain}'
						if attr_val.endswith(domain):
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
				elif attr_type._args['disallowed_domains']:
					for domain in attr_type._args['disallowed_domains']:
						if attr_type._args['strict']:
							domain = f'@{domain}'
						if attr_val.endswith(domain):
							break
					else:
						return return_valid_attr(
							attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
						)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'FILE':
			if type(attr_val) == list and len(attr_val):
				try:
					attr_val = await validate_attr(
						mode='deep',
						attr_name=attr_name,
						attr_type=attr_type,
						attr_val=attr_val[0],
						skip_events=skip_events,
						env=env,
						query=query,
						doc=doc,
						scope=attr_val,
					)
				except Exception as e:
					logger.debug(
						'Exception occurred while validating type \'FILE\'. Exception details:'
					)
					logger.debug(e)
					raise InvalidAttrException(
						attr_name=attr_name,
						attr_type=attr_type,
						val_type=type(attr_val),
					)
			file_type_check = (
				type(attr_val) == dict
				and set(attr_val.keys()) == {'name', 'lastModified', 'type', 'size', 'content'}
				and type(attr_val['name']) == str
				and type(attr_val['type']) == str
				and type(attr_val['lastModified']) == int
				and type(attr_val['size']) == int
				and type(attr_val['content']) in [binary.Binary, bytes]
			)
			if not file_type_check:
				raise InvalidAttrException(
					attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
				)
			if attr_type._args['types']:
				for file_type in attr_type._args['types']:
					if attr_val['type'].split('/')[0] == file_type.split('/')[0]:
						if (
							file_type.split('/')[1] == '*'
							or attr_val['type'].split('/')[1] == file_type.split('/')[1]
						):
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
			else:
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'FLOAT':
			if type(attr_val) == str and re.match(r'^[0-9]+(\.[0-9]+)?$', attr_val):
				attr_val = float(attr_val)
			elif type(attr_val) == int:
				attr_val = float(attr_val)

			if type(attr_val) == float:
				if attr_type._args['ranges']:
					for _range in attr_type._args['ranges']:
						if attr_val >= _range[0] and attr_val < _range[1]:
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'GEO':
			if (
				type(attr_val) == dict
				and set(attr_val.keys()) == {'type', 'coordinates'}
				and attr_val['type'] in ['Point']
				and type(attr_val['coordinates']) == list
				and len(attr_val['coordinates']) == 2
				and type(attr_val['coordinates'][0]) in [int, float]
				and type(attr_val['coordinates'][1]) in [int, float]
			):
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'ID':
			if type(attr_val) == BaseModel or type(attr_val) == DictObj:
				return return_valid_attr(
					attr_val=attr_val._id, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)
			elif type(attr_val) == ObjectId:
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)
			elif type(attr_val) == str:
				try:
					return return_valid_attr(
						attr_val=ObjectId(attr_val), attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)
				except Exception as e:
					logger.debug('Exception occurred while validating type \'ID\'. Exception details:')
					logger.debug(e)
					raise ConvertAttrException(
						attr_name=attr_name,
						attr_type=attr_type,
						val_type=type(attr_val),
					)

		elif attr_type._type == 'INT':
			if type(attr_val) == str and re.match(r'^[0-9]+$', attr_val):
				attr_val = int(attr_val)

			if type(attr_val) == int:
				if attr_type._args['ranges']:
					# [TODO] Implement inf range, ref: https://stackoverflow.com/questions/9884213/looping-from-1-to-infinity-in-python
					for _range in attr_type._args['ranges']:
						if attr_val in range(*_range):
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'IP':
			if re.match(
				r'^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$',
				attr_val,
			):
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'LIST':
			if type(attr_val) == list:
				if attr_type._args['min']:
					if len(attr_val) < attr_type._args['min']:
						raise InvalidAttrException(
							attr_name=attr_name,
							attr_type=attr_type,
							val_type=type(attr_val),
						)
				if attr_type._args['max']:
					if len(attr_val) > attr_type._args['max']:
						raise InvalidAttrException(
							attr_name=attr_name,
							attr_type=attr_type,
							val_type=type(attr_val),
						)
				for i in range(len(attr_val)):
					child_attr_val = attr_val[i]
					child_attr_check = False
					for child_attr_type in attr_type._args['list']:
						try:
							attr_val[i] = await validate_attr(
								mode='deep',
								attr_name=attr_name,
								attr_type=child_attr_type,
								attr_val=child_attr_val,
								skip_events=skip_events,
								env=env,
								query=query,
								doc=doc,
								scope=attr_val,
							)
							child_attr_check = True
							break
						except Exception as e:
							logger.debug(
								'Exception occurred while validating type \'LIST\'. Exception details:'
							)
							logger.debug(e)
					if not child_attr_check:
						raise InvalidAttrException(
							attr_name=attr_name,
							attr_type=attr_type,
							val_type=type(attr_val),
						)
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'LOCALE':
			attr_val = await validate_attr(
				mode='deep',
				attr_name=attr_name,
				attr_type=ATTR.KV_DICT(
					key=ATTR.LITERAL(literal=[locale for locale in Config.locales]),
					val=ATTR.STR(),
					min=1,
					req=[Config.locale],
				),
				attr_val=attr_val,
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				scope=attr_val,
			)
			if Config.locale_strategy == LOCALE_STRATEGY.NONE_VALUE:
				attr_val = {
					locale: attr_val[locale] if locale in attr_val.keys() else None
					for locale in Config.locales
				}
			elif callable(Config.locale_strategy):
				attr_val = {
					locale: attr_val[locale]
					if locale in attr_val.keys()
					else Config.locale_strategy(attr_val=attr_val, locale=locale)
					for locale in Config.locales
				}
			else:
				attr_val = {
					locale: attr_val[locale] if locale in attr_val.keys() else attr_val[Config.locale]
					for locale in Config.locales
				}
			return return_valid_attr(
				attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
			)

		elif attr_type._type == 'LOCALES':
			if attr_val in Config.locales:
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'PHONE':
			if attr_type._args['codes']:
				for phone_code in attr_type._args['codes']:
					if re.match(fr'^\+{phone_code}[0-9]+$', attr_val):
						return return_valid_attr(
							attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
						)
			else:
				if re.match(r'^\+[0-9]+$', attr_val):
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'STR':
			if type(attr_val) == str:
				if attr_type._args['pattern']:
					if re.match(f'^{attr_type._args["pattern"]}$', attr_val):
						return return_valid_attr(
							attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
						)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'TIME':
			if re.match(r'^[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$', attr_val):
				if attr_type._args['ranges']:
					for time_range in attr_type._args['ranges']:
						time_range = copy.deepcopy(time_range)
						for i in [0, 1]:
							if time_range[i][0] in ['+', '-']:
								time_range_delta = {}
								if time_range[i][-1] == 's':
									time_range_delta = {'seconds': int(time_range[i][:-1])}
								elif time_range[i][-1] == 'm':
									time_range_delta = {'minutes': int(time_range[i][:-1])}
								elif time_range[i][-1] == 'h':
									time_range_delta = {'hours': int(time_range[i][:-1])}
								time_range[i] = (
									(datetime.datetime.utcnow() + datetime.timedelta(**time_range_delta))
									.isoformat()
									.split('T')[1]
								)
						if attr_val >= time_range[0] and attr_val < time_range[1]:
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'URI_WEB':
			if re.match(r'^https?:\/\/(?:[\w\-\_]+\.)(?:\.?[\w]{2,})+([\?\/].*)?$', attr_val):
				if attr_type._args['allowed_domains']:
					attr_val_domain = attr_val.split('/')[2]
					for domain in attr_type._args['allowed_domains']:
						if attr_type._args['strict'] and attr_val_domain == domain:
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
						elif not attr_type._args['strict'] and attr_val_domain.endswith(domain):
							return return_valid_attr(
								attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
							)
				elif attr_type._args['disallowed_domains']:
					attr_val_domain = attr_val.split('/')[2]
					for domain in attr_type._args['disallowed_domains']:
						if attr_type._args['strict'] and attr_val_domain == domain:
							break
						elif not attr_type._args['strict'] and attr_val_domain.endswith(domain):
							break
					else:
						return return_valid_attr(
							attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
						)
				else:
					return return_valid_attr(
						attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
					)

		elif attr_type._type == 'LITERAL':
			if attr_val in attr_type._args['literal']:
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'UNION':
			for child_attr in attr_type._args['union']:
				try:
					attr_val = await validate_attr(
						mode='deep',
						attr_name=attr_name,
						attr_type=child_attr,
						attr_val=attr_val,
						skip_events=skip_events,
						env=env,
						query=query,
						doc=doc,
						scope=attr_val,
					)
				except Exception as e:
					logger.debug(
						'Exception occurred while validating type \'UNION\'. Exception details:'
					)
					logger.debug(e)
					continue
				return return_valid_attr(
					attr_val=attr_val, attr_oper=attr_oper, attr_oper_args=attr_oper_args
				)

		elif attr_type._type == 'TYPE':
			try:
				attr_val = await attr_type._args['func'](
					mode=mode,
					attr_name=attr_name,
					attr_type=attr_type,
					attr_val=attr_val,
					skip_events=skip_events,
					env=env,
					query=query,
					doc=doc,
					scope=attr_val,
				)
			except InvalidAttrException as e:
				raise e
			except:
				raise InvalidAttrException(
					attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
				)
			return return_valid_attr(
				attr_val=attr_val,
				attr_oper=attr_oper,
				attr_oper_args=attr_oper_args,
			)

	except InvalidAttrException as e:
		# [DOC] Assign exception to a variable for later processing
		validate_exception = e
	except:
		# [DOC] For any other exception, skip for now to raise
		pass

	if mode == 'update':
		return None
	elif (
		mode in ['create', 'create_draft'] and attr_type._default != NAWAH_VALUES.NONE_VALUE
	):
		return attr_type._default
	else:
		try:
			# [DOC] Attempt to raise InvalidAttrException raised during validation
			raise validate_exception
		except:
			raise InvalidAttrException(
				attr_name=attr_name, attr_type=attr_type, val_type=type(attr_val)
			)


def return_valid_attr(
	*,
	attr_val: Any,
	attr_oper: Literal[
		None, '$add', '$multiply', '$append', '$set_index', '$del_val', '$del_index'
	],
	attr_oper_args: Dict[str, Any],
) -> Any:
	if not attr_oper:
		return attr_val
	elif attr_oper in ['$add', '$multiply']:
		return {attr_oper: attr_val, '$field': attr_oper_args['$field']}
	elif attr_oper == '$del_val':
		return {attr_oper: attr_val}
	elif attr_oper == '$append':
		return {'$append': attr_val[0], '$unique': attr_oper_args['$unique']}
	elif attr_oper == '$set_index':
		return {'$set_index': attr_val[0], '$index': attr_oper_args['$index']}
	elif attr_oper == '$del_index':
		return {'$del_index': attr_oper_args['$index']}


def generate_dynamic_attr(
	*, dynamic_attr: Dict[str, Any]
) -> Tuple[ATTR, Dict[str, Any]]:
	# [DOC] Fail-safe checks
	if dynamic_attr['type'] not in ATTRS_TYPES_ARGS.keys():
		raise InvalidAttrTypeException(attr_type=dynamic_attr['type'])
	if 'args' not in dynamic_attr.keys():
		dynamic_attr['args'] = {}

	# [DOC] Process args of type ATTR
	if dynamic_attr['type'] == 'LIST':
		shadow_arg_list: List[Optional[Dict[str, Any]]] = []
		for i in range(len(dynamic_attr['args']['list'])):
			shadow_arg_list.append(None)
			dynamic_attr['args']['list'][i], shadow_arg_list[i] = generate_dynamic_attr(
				dynamic_attr=dynamic_attr['args']['list'][i]
			)
	elif dynamic_attr['type'] == 'TYPED_DICT':
		shadow_arg_dict = {}
		for dict_attr in dynamic_attr['args']['dict'].keys():
			(
				dynamic_attr['args']['dict'][dict_attr],
				shadow_arg_dict[dict_attr],
			) = generate_dynamic_attr(
				dynamic_attr=dynamic_attr['args']['dict'][dict_attr]
			)
	elif dynamic_attr['type'] == 'KV_DICT':
		dynamic_attr['args']['key'], _ = generate_dynamic_attr(
			dynamic_attr=dynamic_attr['args']['key']
		)
		dynamic_attr['args']['val'], _ = generate_dynamic_attr(
			dynamic_attr=dynamic_attr['args']['val']
		)
	if dynamic_attr['type'] == 'UNION':
		shadow_arg_union: List[Optional[Dict[str, Any]]] = []
		for i in range(len(dynamic_attr['args']['union'])):
			shadow_arg_list.append(None)
			dynamic_attr['args']['union'][i], shadow_arg_union[i] = generate_dynamic_attr(
				dynamic_attr=dynamic_attr['args']['union'][i]
			)
	# [DOC] Generate dynamic ATTR using ATTR controller
	dynamic_attr_type = getattr(ATTR, dynamic_attr['type'])(**dynamic_attr['args'])
	# [DOC] Reset values for args of type ATTR
	if dynamic_attr['type'] == 'LIST':
		dynamic_attr['args']['list'] = shadow_arg_list
	elif dynamic_attr['type'] == 'TYPED_DICT':
		dynamic_attr['args']['dict'] = shadow_arg_dict
	elif dynamic_attr['type'] == 'UNION':
		dynamic_attr['args']['dict'] = shadow_arg_union
	# [DOC] Set defaults for optional args
	if 'allow_none' not in dynamic_attr.keys():
		dynamic_attr['allow_none'] = False
	if 'default' not in dynamic_attr.keys():
		dynamic_attr['default'] = None

	# [DOC] Validate Attr Type for possible Attr Type TYPE presence, to set ATTR._args[func] value
	ATTR.validate_type(attr_type=dynamic_attr_type)

	return (dynamic_attr_type, dynamic_attr)

def _set_none_value(
	*,
	doc: Union[NAWAH_DOC, List[NAWAH_DOC]],
):
	if type(doc) == dict:
		doc_iter = doc.keys()
	elif type(doc) == list:
		doc_iter = range(len(doc))

	for j in doc_iter:
		if type(doc[j]) in [list, dict]:
			_set_none_value(doc=doc[j])
		elif doc[j] == NAWAH_VALUES.NONE_VALUE:
			doc[j] = None




from nawah.enums import NAWAH_VALUES
from nawah.config import Config

from bson import ObjectId
from typing import (
	Dict,
	Union,
	Type,
	List,
	Sequence,
	Any,
	Literal,
	Callable,
	Optional,
	ForwardRef,
	TYPE_CHECKING,
	Protocol,
)

import datetime, logging, re, inspect

from ._package import SYS_DOC
from ._exceptions import (
	InvalidAttrTypeException,
	InvalidAttrTypeArgException,
	InvalidAttrTypeArgsException,
)

if TYPE_CHECKING:
	from ._module import EXTN
	from ._types import NAWAH_EVENTS, NAWAH_ENV, NAWAH_QUERY, NAWAH_DOC, NAWAH_DOC
	from ._query import Query

logger = logging.getLogger('nawah')

ATTRS_TYPES_ARGS: Dict[str, Dict[str, Union[Type, str]]] = {
	'ANY': {},
	'ACCESS': {},
	'COUNTER': {'pattern': str},
	'ID': {},
	'STR': {'pattern': str},
	'INT': {'ranges': List[List[int]]},
	'FLOAT': {'ranges': List[List[float]]},
	'BOOL': {},
	'LOCALE': {},
	'LOCALES': {},
	'EMAIL': {
		'allowed_domains': List[str],
		'disallowed_domains': List[str],
		'strict_matching': bool,
	},
	'PHONE': {'codes': List[str]},
	'IP': {},
	'URI_WEB': {
		'allowed_domains': List[str],
		'disallowed_domains': List[str],
		'strict_matching': bool,
	},
	'DATETIME': {'ranges': List[List[datetime.datetime]]},
	'DATE': {'ranges': List[List[datetime.date]]},
	'DYNAMIC_ATTR': {'types': List[str]},
	'DYNAMIC_VAL': {'dynamic_attr': str},
	'TIME': {'ranges': List[List[datetime.time]]},
	'FILE': {
		'types': List[str],
		'max_ratio': List[int],
		'min_ratio': List[int],
		'max_dims': List[int],
		'min_dims': List[int],
		'max_size': int,
	},
	'GEO': {},
	'LIST': {'list': List['ATTR'], 'min': int, 'max': int},
	'KV_DICT': {
		'key': 'ATTR',
		'val': 'ATTR',
		'min': int,
		'max': int,
		'req': List[str],
	},
	'TYPED_DICT': {'dict': Dict[str, 'ATTR']},
	'LITERAL': {'literal': List[Union[str, int, float, bool]]},
	'UNION': {'union': List['ATTR']},
	'TYPE': {'type': str},
}

SPECIAL_ATTRS = Literal[
	'$search', '$sort', '$skip', '$limit', '$extn', '$soft', '$attrs', '$group', '$geo_near'
]


class ATTR_TYPE_CALLABLE_TYPE(Protocol):
	def __call__(
		*,
		self,
		mode: Literal['create', 'create_draft', 'update'],
		attr_name: str,
		attr_type: 'ATTR',
		attr_val: Any,
		skip_events: 'NAWAH_EVENTS',
		env: 'NAWAH_ENV',
		query: Union['NAWAH_QUERY', 'Query'],
		doc: 'NAWAH_DOC',
		scope: Optional['NAWAH_DOC'],
	) -> Any:
		...


ATTRS_TYPES_TYPE = Literal[
	'ANY',
	'ACCESS',
	'COUNTER',
	'ID',
	'STR',
	'INT',
	'FLOAT',
	'BOOL',
	'LOCALE',
	'LOCALES',
	'EMAIL',
	'PHONE',
	'IP',
	'URI_WEB',
	'DATETIME',
	'DATE',
	'DYNAMIC_ATTR',
	'DYNAMIC_VAL',
	'TIME',
	'FILE',
	'GEO',
	'LIST',
	'KV_DICT',
	'TYPED_DICT',
	'LITERAL',
	'UNION',
	'TYPE',
]


class ATTR:
	_nawah_attr: bool = True
	_type: ATTRS_TYPES_TYPE
	_desc: Optional[str]
	_args: Dict[str, Any]
	_valid: bool = False
	_extn: Optional[Union['EXTN', 'ATTR']] = None

	__default = NAWAH_VALUES.NONE_VALUE

	@property
	def _default(self):
		if self.__default == '$__datetime':
			return datetime.datetime.utcnow().isoformat()
		elif self.__default == '$__date':
			return datetime.date.today().isoformat()
		elif self.__default == '$__time':
			return datetime.datetime.now().time().isoformat()
		else:
			return self.__default

	@_default.setter
	def _default(self, value):
		self.__default = value

	def __repr__(self):
		return f'<ATTR:{self._type},{self._args}>'

	def __init__(self, *, attr_type: ATTRS_TYPES_TYPE, desc: str = None, **kwargs: Any):
		self._type = attr_type
		self._desc = desc
		self._args = kwargs
		ATTR.validate_type(attr_type=self, skip_type=True)

	@classmethod
	def ANY(cls, *, desc: str = None):
		return ATTR(attr_type='ANY', desc=desc)

	@classmethod
	def ACCESS(cls, *, desc: str = None):
		return ATTR(attr_type='ACCESS', desc=desc)

	@classmethod
	def COUNTER(cls, *, desc: str = None, pattern: str, values: List[Callable] = None):
		return ATTR(attr_type='COUNTER', desc=desc, pattern=pattern, values=values)

	@classmethod
	def ID(cls, *, desc: str = None):
		return ATTR(attr_type='ID', desc=desc)

	@classmethod
	def STR(cls, *, desc: str = None, pattern: str = None):
		return ATTR(attr_type='STR', desc=desc, pattern=pattern)

	@classmethod
	def INT(cls, *, desc: str = None, ranges: List[List[int]] = None):
		return ATTR(attr_type='INT', desc=desc, ranges=ranges)

	@classmethod
	def FLOAT(cls, *, desc: str = None, ranges: List[List[int]] = None):
		return ATTR(attr_type='FLOAT', desc=desc, ranges=ranges)

	@classmethod
	def BOOL(cls, *, desc: str = None):
		return ATTR(attr_type='BOOL', desc=desc)

	@classmethod
	def LOCALE(cls, *, desc: str = None):
		return ATTR(attr_type='LOCALE', desc=desc)

	@classmethod
	def LOCALES(cls, *, desc: str = None):
		return ATTR(attr_type='LOCALES', desc=desc)

	@classmethod
	def EMAIL(
		cls,
		*,
		desc: str = None,
		allowed_domains: List[str] = None,
		disallowed_domains: List[str] = None,
		strict: bool = False,
	):
		return ATTR(
			attr_type='EMAIL',
			desc=desc,
			allowed_domains=allowed_domains,
			disallowed_domains=disallowed_domains,
			strict=strict,
		)

	@classmethod
	def PHONE(cls, *, desc: str = None, codes: List[str] = None):
		return ATTR(attr_type='PHONE', desc=desc, codes=codes)

	@classmethod
	def IP(cls, *, desc: str = None):
		return ATTR(attr_type='IP', desc=desc)

	@classmethod
	def URI_WEB(
		cls,
		*,
		desc: str = None,
		allowed_domains: List[str] = None,
		disallowed_domains: List[str] = None,
		strict: bool = False,
	):
		return ATTR(
			attr_type='URI_WEB',
			desc=desc,
			allowed_domains=allowed_domains,
			disallowed_domains=disallowed_domains,
			strict=strict,
		)

	@classmethod
	def DATETIME(cls, *, desc: str = None, ranges: List[List[str]] = None):
		return ATTR(attr_type='DATETIME', desc=desc, ranges=ranges)

	@classmethod
	def DATE(cls, *, desc: str = None, ranges: List[List[str]] = None):
		return ATTR(attr_type='DATE', desc=desc, ranges=ranges)

	@classmethod
	def DYNAMIC_ATTR(cls, *, desc: str = None, types: List[str] = None):
		return ATTR(attr_type='DYNAMIC_ATTR', desc=desc, types=types)

	@classmethod
	def DYNAMIC_VAL(cls, *, desc: str = None, dynamic_attr: str):
		return ATTR(attr_type='DYNAMIC_VAL', desc=desc, dynamic_attr=dynamic_attr)

	@classmethod
	def TIME(cls, *, desc: str = None, ranges: List[List[str]] = None):
		return ATTR(attr_type='TIME', desc=desc, ranges=ranges)

	@classmethod
	def FILE(
		cls,
		*,
		desc: str = None,
		types: List[str] = None,
		max_ratio: List[int] = None,
		min_ratio: List[int] = None,
		max_dims: List[int] = None,
		min_dims: List[int] = None,
		max_size: int = None,
	):
		return ATTR(
			attr_type='FILE',
			desc=desc,
			types=types,
			max_ratio=max_ratio,
			min_ratio=min_ratio,
			max_dims=max_dims,
			min_dims=min_dims,
			max_size=max_size,
		)

	@classmethod
	def GEO(cls, *, desc: str = None):
		return ATTR(attr_type='GEO', desc=desc)

	@classmethod
	def LIST(
		cls,
		*,
		desc: str = None,
		list: List['ATTR'],
		min: int = None,
		max: int = None,
	):
		return ATTR(attr_type='LIST', desc=desc, list=list, min=min, max=max)

	@classmethod
	def KV_DICT(
		cls,
		*,
		desc: str = None,
		key: 'ATTR',
		val: 'ATTR',
		min: int = None,
		max: int = None,
		req: List[str] = None,
	):
		return ATTR(
			attr_type='KV_DICT', desc=desc, key=key, val=val, min=min, max=max, req=req
		)

	@classmethod
	def TYPED_DICT(cls, *, desc: str = None, dict: Dict[str, 'ATTR']):
		return ATTR(attr_type='TYPED_DICT', desc=desc, dict=dict)

	@classmethod
	def LITERAL(cls, *, desc: str = None, literal: Sequence[Union[str, int, float, bool]]):
		return ATTR(attr_type='LITERAL', desc=desc, literal=literal)

	@classmethod
	def UNION(cls, *, desc: str = None, union: List['ATTR']):
		return ATTR(attr_type='UNION', desc=desc, union=union)

	@classmethod
	def TYPE(cls, *, desc: str = None, type: Union[str, ATTR_TYPE_CALLABLE_TYPE]):
		return ATTR(attr_type='TYPE', desc=desc, type=type)

	@classmethod
	def validate_type(cls, *, attr_type: 'ATTR', skip_type: bool = False):
		# [DOC] Skip validating Attr Type if it is already validated, unless we want to validate TYPE, which could require deep (nested) validation
		if skip_type and attr_type._valid:
			return

		if attr_type._type not in ATTRS_TYPES_ARGS.keys():
			raise InvalidAttrTypeException(attr_type=attr_type)
		elif not skip_type and attr_type._type == 'TYPE':
			# [DOC] Complex condition is unreadblae in single condition set, break into multiple conditions as try..except block
			try:
				if type(attr_type._args['type']) == str:
					if attr_type._args['type'] not in Config.types.keys():
						raise Exception()
					if not inspect.iscoroutinefunction(Config.types[attr_type._args['type']]):
						raise Exception()
					# [DOC] Assign new Attr Type Arg for shorthand calling the TYPE function
					attr_type._args['func'] = Config.types[attr_type._args['type']]
				else:
					if not inspect.iscoroutinefunction(attr_type._args['type']):
						raise Exception()
					# [DOC] Assign new Attr Type Arg for shorthand calling the TYPE function
					attr_type._args['func'] = attr_type._args['type']
			except:
				raise InvalidAttrTypeException(attr_type=attr_type)
		elif attr_type._type != 'TYPE':
			for arg in ATTRS_TYPES_ARGS[attr_type._type].keys():
				if (
					arg in ['list', 'dict', 'literal', 'union', 'type']
					and arg not in attr_type._args.keys()
				):
					raise InvalidAttrTypeArgException(
						arg_name=arg,
						arg_type=ATTRS_TYPES_ARGS[attr_type._type][arg],
						arg_val=attr_type._args[arg],
					)
				elif arg in attr_type._args.keys() and attr_type._args[arg] != None:
					cls.validate_arg(
						arg_name=arg,
						arg_type=ATTRS_TYPES_ARGS[attr_type._type][arg],
						arg_val=attr_type._args[arg],
						skip_type=skip_type,
					)
				else:
					attr_type._args[arg] = None
			if attr_type._type == 'COUNTER':
				if (
					'$__values.' in attr_type._args['pattern']
					or '$__counters:' in attr_type._args['pattern']
				):
					logger.error(
						'Attr Type COUNTER is using wrong format for \'$__values\', or \'$__counters\'.'
					)
					raise InvalidAttrTypeException(attr_type=attr_type)
				counter_groups = re.findall(
					r'(\$__(?:values:[0-9]+|counters\.[a-z0-9_]+))',
					attr_type._args['pattern'],
				)
				if len(counter_groups) == 0:
					logger.error(
						'Attr Type COUNTER is not having any \'$__values\', or \'$__counters\'.'
					)
					raise InvalidAttrTypeException(attr_type=attr_type)
				if '$__counters.' not in attr_type._args['pattern']:
					logger.warning('Attr Type COUNTER is defined with not \'$__counters\'.')
				for group in counter_groups:
					if group.startswith('$__counters.'):
						Config.docs.append(
							SYS_DOC(
								module='setting',
								key='var',
								doc={
									'user': ObjectId('f00000000000000000000010'),
									'var': '__counter:' + group.replace('$__counters.', ''),
									'val_type': {'type': 'INT', 'args': {}, 'allow_none': False, 'default': None},
									'val': 0,
									'type': 'global',
								},
							)
						)
			attr_type._valid = True

	@classmethod
	def validate_arg(cls, *, arg_name: str, arg_type: Any, arg_val: Any, skip_type: bool):
		if arg_type == str:
			if type(arg_val) != str:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			return
		elif arg_type == int:
			if type(arg_val) != int:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			return
		elif arg_type == float:
			if type(arg_val) not in [float, int]:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			return
		elif arg_type == bool:
			if type(arg_val) != bool:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			return
		elif type(arg_type) == ForwardRef or (type(arg_type) == str and arg_type == 'ATTR'):
			if type(arg_val) != ATTR:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			if arg_name == 'key':
				if arg_val._type not in ['STR', 'LITERAL']:
					raise InvalidAttrTypeArgException(
						arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
					)
			return
		elif arg_name == 'literal':
			if type(arg_val) != list:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			for arg_val_child in arg_val:
				if type(arg_val_child) not in arg_type.__args__[0].__args__:
					raise InvalidAttrTypeArgException(
						arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
					)
			return
		elif arg_name == 'union':
			if type(arg_val) != list:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			for arg_val_child in arg_val:
				if type(arg_val_child) != ATTR:
					raise InvalidAttrTypeArgException(
						arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
					)
				ATTR.validate_type(attr_type=arg_val_child, skip_type=skip_type)
			return
		elif arg_type == datetime.date:
			if not re.match(r'^[0-9]{4}-[0-9]{2}-[0-9]{2}$', arg_val) and not re.match(
				r'^[\-\+][0-9]+[dsmhw]$', arg_val
			):
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			return
		elif arg_type == datetime.datetime:
			if (
				not re.match(
					r'^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$',
					arg_val,
				)
				and not re.match(r'^[\-\+][0-9]+[dsmhw]$', arg_val)
			):
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			return
		elif arg_type == datetime.time:
			if not re.match(
				r'^[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?$', arg_val
			) and not re.match(r'^[\-\+][0-9]+[dsmhw]$', arg_val):
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			return
		elif arg_type._name == 'List':
			if type(arg_val) != list:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			for arg_val_child in arg_val:
				cls.validate_arg(
					arg_name=arg_name,
					arg_type=arg_type.__args__[0],
					arg_val=arg_val_child,
					skip_type=skip_type,
				)

				# [DOC] In Addition to validating as list arg, check for ATTR and validate
				if type(arg_val_child) == ATTR:
					ATTR.validate_type(attr_type=arg_val_child, skip_type=skip_type)

			return
		elif arg_type._name == 'Dict':
			if type(arg_val) != dict:
				raise InvalidAttrTypeArgException(
					arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
				)
			for arg_val_child in arg_val.keys():
				if type(arg_val[arg_val_child]) != ATTR:
					raise InvalidAttrTypeArgException(
						arg_name=arg_name, arg_type=arg_type, arg_val=arg_val[arg_val_child]
					)
				ATTR.validate_type(attr_type=arg_val[arg_val_child], skip_type=skip_type)
			return

		raise InvalidAttrTypeArgException(
			arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
		)



from nawah.registry import Registry
from nawah.classes import (
	NAWAH_ENV,
	NAWAH_QUERY,
	Query,
	NAWAH_DOC,
	PERM,
	ATTR,
	InvalidPermissionsExcpetion,
	InvalidAttrException,
)
from nawah.utils import _extract_attr, validate_attr
from nawah.enums import Event

from typing import List, Dict, Union, Any, Iterable, TYPE_CHECKING

import copy, datetime, logging

if TYPE_CHECKING:
	from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


async def _check_permissions(
	skip_events: List[Event],
	env: NAWAH_ENV,
	query: Union[NAWAH_QUERY, Query],
	doc: NAWAH_DOC,
	module: 'BaseModule',
	permissions: List[PERM],
):
	user = env['session'].user

	permissions = copy.deepcopy(permissions)

	for permission in permissions:
		logger.debug(f'checking permission: {permission} against: {user.privileges}')
		permission_pass = False
		if permission.privilege == '*':
			permission_pass = True

		if not permission_pass:
			if permission.privilege.find('.') == -1:
				permission_module = module.module_name
				permission_attr = permission.privilege
			elif permission.privilege.find('.') != -1:
				permission_module = permission.privilege.split('.')[0]
				permission_attr = permission.privilege.split('.')[1]

			if '*' in user.privileges.keys():
				user.privileges[permission_module] = copy.deepcopy(user.privileges['*'])
			if permission_module in user.privileges.keys():
				if (
					type(user.privileges[permission_module]) == list
					and '*' in user.privileges[permission_module]
				):
					user.privileges[permission_module] += copy.deepcopy(
						Registry.module(permission_module).privileges
					)
			if permission_module not in user.privileges.keys():
				user.privileges[permission_module] = []

			if permission_attr in user.privileges[permission_module]:
				permission_pass = True

		if permission_pass:
			query_mod = await _parse_permission_args(
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				permission_args=permission.query_mod,
			)
			doc_mod = await _parse_permission_args(
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				permission_args=permission.doc_mod,
			)

			return {'query_mod': query_mod, 'doc_mod': doc_mod}
	# [DOC] If all permission checks fail
	raise InvalidPermissionsExcpetion()


async def _parse_permission_args(
	skip_events: List[Event],
	env: NAWAH_ENV,
	query: Union[NAWAH_QUERY, Query],
	doc: NAWAH_DOC,
	permission_args: Any,
):
	user = env['session'].user

	args_iter: Iterable

	if type(permission_args) == list:
		args_iter = range(len(permission_args))
	elif type(permission_args) == dict:
		args_iter = list(permission_args.keys())

	for j in args_iter:
		if type(permission_args[j]) == ATTR:
			try:
				permission_args[j] = await validate_attr(
					mode='create',
					attr_name=j,
					attr_type=permission_args[j],
					attr_val=doc[j],
					skip_events=[],
					env=env,
					query=query,
					doc=doc,
					scope=doc,
				)
			except InvalidAttrException as e:
				raise e
			except:
				# [DOC] There is a chance doc[j] is invalid, so try to get the type in try..except
				try:
					raise InvalidAttrException(
						attr_name=j,
						attr_type=permission_args[j],
						val_type=type(doc[j]),
					)
				except:
					raise InvalidAttrException(
						attr_name=j,
						attr_type=permission_args[j],
						val_type=None,
					)
		elif type(permission_args[j]) == dict:
			# [DOC] Check opers
			for oper in [
				'$gt',
				'$lt',
				'$gte',
				'$lte',
				'$bet',
				'$ne',
				'$regex',
				'$all',
				'$in',
				'$nin',
			]:
				if oper in permission_args[j].keys():
					if oper == '$bet':
						permission_args[j]['$bet'] = await _parse_permission_args(
							skip_events=skip_events,
							env=env,
							query=query,
							doc=doc,
							permission_args=permission_args[j]['$bet'],
						)
					else:
						permission_args[j][oper] = await _parse_permission_args(
							skip_events=skip_events,
							env=env,
							query=query,
							doc=doc,
							permission_args={oper: permission_args[j][oper]},
						)
						permission_args[j][oper] = permission_args[j][oper][oper]
					# [DOC] Continue the iteration
					continue
			# [DOC] Child args, parse
			permission_args[j] = await _parse_permission_args(
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				permission_args=permission_args[j],
			)
		elif type(permission_args[j]) == list:
			permission_args[j] = await _parse_permission_args(
				skip_events=skip_events,
				env=env,
				query=query,
				doc=doc,
				permission_args=permission_args[j],
			)
		elif type(permission_args[j]) == str:
			# [DOC] Check for variables
			if permission_args[j] == '$__user':
				permission_args[j] = user._id
			elif permission_args[j].startswith('$__user.'):
				try:
					permission_args[j] = _extract_attr(
						scope=user,
						attr_path=permission_args[j].replace('$__user.', '$__'),
					)
				except Exception as e:
					# [TODO] Log exception
					# [DOC] For values that are expected to have a list value, return empty list
					if type(permission_args) == dict and j in ['$in', '$nin']:
						permission_args[j] = [None]
					else:
						permission_args[j] = None
			elif permission_args[j] == '$__access':
				permission_args[j] = {'$__user': user._id, '$__groups': user.groups}
			elif permission_args[j] == '$__datetime':
				permission_args[j] = datetime.datetime.utcnow().isoformat()
			elif permission_args[j] == '$__date':
				permission_args[j] = datetime.date.today().isoformat()
			elif permission_args[j] == '$__time':
				permission_args[j] = datetime.datetime.now().time().isoformat()

	return permission_args



from nawah.classes import ATTR

from typing import Union, List, Dict, cast, Any, Literal, MutableMapping

import logging

logger = logging.getLogger('nawah')


def _extract_attr(*, scope: MutableMapping[str, Any], attr_path: str):
	if attr_path.startswith('$__'):
		attr_path_parts = attr_path[3:].split('.')
	else:
		attr_path_parts = attr_path.split('.')
	attr: Any = scope
	for i in range(len(attr_path_parts)):
		child_attr = attr_path_parts[i]
		try:
			logger.debug(f'Attempting to extract {child_attr} from {attr}.')
			if ':' in child_attr:
				child_attr_parts = child_attr.split(':')
				attr = attr[child_attr_parts[0]]
				for i in range(1, len(child_attr_parts)):
					attr = attr[int(child_attr_parts[i])]  # type: ignore
			else:
				attr = attr[child_attr]
		except Exception as e:
			logger.error(f'Failed to extract {child_attr} from {attr}.')
			raise e
	return attr


def _set_attr(*, scope: Dict[str, Any], attr_path: str, value: Any):
	if attr_path.startswith('$__'):
		attr_path_parts = attr_path[3:].split('.')
	else:
		attr_path_parts = attr_path.split('.')
	attr = scope
	for i in range(len(attr_path_parts) - 1):
		child_attr = attr_path_parts[i]
		try:
			if ':' in child_attr:
				child_attr_parts = child_attr.split(':')
				attr = attr[child_attr_parts[0]]
				for i in range(1, len(child_attr_parts)):
					attr = attr[int(child_attr_parts[i])]  # type: ignore
			else:
				attr = attr[child_attr]
		except Exception as e:
			logger.error(f'Failed to extract {child_attr} from {attr}.')
			raise e
	if ':' in attr_path_parts[-1]:
		attr_path_parts_last = attr_path_parts[-1].split(':')
		attr = attr[attr_path_parts_last[0]]
		for i in range(1, len(attr_path_parts_last) - 1):
			attr = attr[int(attr_path_parts_last[i])]  # type: ignore
		attr[int(attr_path_parts_last[-1])] = value  # type: ignore
	else:
		attr[attr_path_parts[-1]] = value


def _expand_attr(*, doc: Dict[str, Any], expanded_doc: Dict[str, Any] = None):
	if not expanded_doc:
		expanded_doc = {}
	for attr in doc.keys():
		if type(doc[attr]) == dict:
			doc[attr] = _expand_attr(doc=doc[attr])
		if '.' in attr:
			attr_path = attr.split('.')
			scope = expanded_doc
			for i in range(len(attr_path) - 1):
				try:
					if type(scope[attr_path[i]]) != dict:
						scope[attr_path[i]] = {}
				except KeyError:
					scope[attr_path[i]] = {}
				scope = scope[attr_path[i]]
			scope[attr_path[-1]] = doc[attr]
		else:
			expanded_doc[attr] = doc[attr]
	return expanded_doc


def _deep_update(*, target: Union[List, Dict], new_values: Union[List, Dict]):
	if type(target) != type(new_values):
		logger.error(
			f'Type \'{type(target)}\' of \'target\' is not the same as \'{type(new_values)}\' of \'new_values\'. Exiting.'
		)
		exit(1)
	if type(new_values) == dict:
		new_values = cast(dict, new_values)
		for k in new_values.keys():
			target = cast(dict, target)
			if k not in target.keys():
				target[k] = new_values[k]
			else:
				_deep_update(target=target[k], new_values=new_values[k])
	elif type(new_values) == list:
		for j in new_values:
			target = cast(list, target)
			if j not in target:
				target.append(j)


def _update_attr_values(
	*, attr: ATTR, value: Literal['default', 'extn'], value_path: str, value_val: Any
):

	value_path_part = value_path.split('.')
	for child_default_path in value_path_part:
		if ':' in child_default_path:
			attr = attr._args['dict'][child_default_path.split(':')[0]]._args['list'][
				int(child_default_path.split(':')[1])
			]
		else:
			if child_default_path == 'val' and attr._type == 'KV_DICT':
				attr = attr._args['val']
			else:
				attr = attr._args['dict'][child_default_path]
	setattr(attr, f'_{value}', value_val)



from nawah.config import Config
from nawah.enums import DELETE_STRATEGY
from nawah.classes import NAWAH_ENV, ATTR, UnknownDeleteStrategyException

from bson import ObjectId
from typing import Dict, List, Any, Union

import logging

logger = logging.getLogger('nawah')


async def delete(
	*,
	env: NAWAH_ENV,
	collection_name: str,
	attrs: Dict[str, ATTR],
	docs: List[Union[str, ObjectId]],
	strategy: DELETE_STRATEGY,
) -> Dict[str, Any]:
	# [DOC] Check strategy to cherrypick update, delete calls and system_docs
	if strategy in [DELETE_STRATEGY.SOFT_SKIP_SYS, DELETE_STRATEGY.SOFT_SYS]:
		if strategy == DELETE_STRATEGY.SOFT_SKIP_SYS:
			del_docs = [
				ObjectId(doc) for doc in docs if ObjectId(doc) not in Config._sys_docs.keys()
			]
			if len(del_docs) != len(docs):
				logger.warning(
					'Skipped soft delete for system docs due to \'DELETE_SOFT_SKIP_SYS\' strategy.'
				)
		else:
			logger.warning('Detected \'DELETE_SOFT_SYS\' strategy for delete call.')
			del_docs = [ObjectId(doc) for doc in docs]
		# [DOC] Perform update call on matching docs
		collection = env['conn'][Config.data_name][collection_name]
		update_doc = {'$set': {'__deleted': True}}
		# [DOC] If using Azure Mongo service update docs one by one
		if Config.data_azure_mongo:
			update_count = 0
			for _id in docs:
				results = await collection.update_one({'_id': _id}, update_doc)
				update_count += results.modified_count
		else:
			results = await collection.update_many({'_id': {'$in': docs}}, update_doc)
			update_count = results.modified_count
		return {'count': update_count, 'docs': [{'_id': doc} for doc in docs]}
	elif strategy in [DELETE_STRATEGY.FORCE_SKIP_SYS, DELETE_STRATEGY.FORCE_SYS]:
		if strategy == DELETE_STRATEGY.FORCE_SKIP_SYS:
			del_docs = [
				ObjectId(doc) for doc in docs if ObjectId(doc) not in Config._sys_docs.keys()
			]
			if len(del_docs) != len(docs):
				logger.warning(
					'Skipped soft delete for system docs due to \'DELETE_FORCE_SKIP_SYS\' strategy.'
				)
		else:
			logger.warning('Detected \'DELETE_FORCE_SYS\' strategy for delete call.')
			del_docs = [ObjectId(doc) for doc in docs]
		# [DOC] Perform delete query on matching docs
		collection = env['conn'][Config.data_name][collection_name]
		if Config.data_azure_mongo:
			delete_count = 0
			for _id in del_docs:
				results = await collection.delete_one({'_id': _id})
				delete_count += results.deleted_count
		else:
			results = await collection.delete_many({'_id': {'$in': del_docs}})
			delete_count = results.deleted_count
		return {'count': delete_count, 'docs': [{'_id': doc} for doc in docs]}
	else:
		raise UnknownDeleteStrategyException(f'DELETE_STRATEGY \'{strategy}\' is unknown.')



from nawah import __version__

from typing import Dict, Any, Optional

import argparse, os, sys, logging, json, subprocess, shutil, pkgutil, traceback

logger = logging.getLogger('nawah')


def packages_install(args: argparse.Namespace):
	app_path = os.path.realpath('.')
	sys.path.insert(0, app_path)

	with open(os.path.join(app_path, 'packages.json')) as f:
		packages = json.loads(f.read())

	for package_name in packages:
		logger.info(f'Attempting to install package \'{package_name}\'')
		_packages_add(
			package_name=package_name,
			source=packages[package_name]['source'],
			version=packages[package_name]['version'],
			auth=packages[package_name]['auth']
			if 'auth' in packages[package_name].keys()
			else None,
		)

	logger.info('Done installing all packages!')


# [DOC] The single underscore to indicate this is not being called directly by argsparser but through a proxy callable
def _packages_add(*, package_name: str, source: str, version: str, auth: Optional[str]):
	from . import _set_testing
	import json

	_set_testing(True)
	logger.info('Checking packages conflicts.')
	app_path = os.path.realpath(os.path.join('.'))
	packages_path = os.path.realpath(os.path.join('.', 'packages'))
	package_path = os.path.realpath(os.path.join('.', 'packages', package_name))

	with open(os.path.join(app_path, 'packages.json')) as f:
		packages = json.loads(f.read())

	api_level = '.'.join(__version__.split('.')[:2])

	if os.path.exists(package_path):
		logger.info(
			f'Package \'{package_name}\' already exists in app. Attempting to test compatibility with API Level and version.'
		)
		try:
			sys.path.insert(0, os.path.realpath('.'))
			sys.path.insert(0, packages_path)
			package = __import__(package_name)
		except:
			logger.error('Failed to load package to test compatibility and version. Exiting.')
			exit(1)

		if package.config.api_level != api_level:
			logger.error(
				f'App is using API Level {api_level}, but package is on API Level {package.config.api_level}. Exiting.'
			)
			exit(1)

		logger.info(f'Package is compatible with app API Level. Attempting to check version.')

		if package.config.version != (package_version := packages[package_name]['version']):
			logger.error(
				f'Package is on version {package.config.version}, but app requires version {package_version}. Exiting.'
			)
			exit(1)

		logger.info(f'Package is already installed with correct version.')

		_set_testing(False)

		return

	authed_source = source
	try:
		if auth:
			logger.info(
				f'Detected \'auth\' configuration for source. Attempting to authorise source.'
			)
			auth_username, auth_password = auth.split(':')
			if auth_username.startswith('__env.'):
				logger.info(
					f'Detected environement variable for \'auth_username\' configuration. Setting it.'
				)
				auth_username = os.environ[auth_username.replace('__env.', '')]
			if auth_password.startswith('__env.'):
				logger.info(
					f'Detected environement variable for \'auth_password\' configuration. Setting it.'
				)
				auth_password = os.environ[auth_password.replace('__env.', '')]
			authed_source = (
				authed_source.split('://', 1)[0]
				+ '://'
				+ auth_username
				+ ':'
				+ auth_password
				+ '@'
				+ authed_source.split('://', 1)[1]
			)
			logger.info(f'Processed \'auth\' configuration successfully.')
	except:
		logger.error(f'Failed to process \'auth\' configuration. Exiting.')
		exit(1)

	logger.info(f'Attempting to add package \'{package_name}\' from source \'{source}\'.')

	pip_command = [
		sys.executable,
		'-m',
		'pip',
		'install',
		'--no-deps',
		'--target',
		packages_path,
		'--extra-index-url',
		authed_source,
		package_name if version == 'latest' else f'{package_name}=={version}',
	]

	pip_call = subprocess.call(pip_command)

	if pip_call != 0:
		logger.error('Last \'pip\' call failed. Check console for more details. Exiting.')
		exit(1)

	logger.info(
		f'Package installed. Attempting to test compatibility with API Level {api_level}.'
	)

	try:
		sys.path.insert(0, os.path.realpath('.'))
		sys.path.insert(0, packages_path)
		package = __import__(package_name)
	except Exception as e:
		logger.error(traceback.format_exc())
		logger.error('Failed to load package to test compatibility. Exiting.')
		exit(1)

	if package.config.api_level != api_level:
		logger.error(
			f'App is using API Level {api_level}, but package is on API Level {package.config.api_level}.'
		)
		_packages_rm(package_name=package_name, confirm=False)

	logger.info(
		f'Package is compatible with app API Level. Attempting to install package requirements.'
	)

	requirements_path = os.path.join(package_path, 'requirements.txt')

	pip_command = [
		sys.executable,
		'-m',
		'pip',
		'install',
		'--user',
		'-r',
		requirements_path,
	]

	pip_call = subprocess.call(pip_command)

	if pip_call != 0:
		logger.error('Last \'pip\' call failed. Check console for more details.')

	logger.info('Attempting to update \'packages.json\'.')

	with open(os.path.realpath(os.path.join('.', 'packages.json')), 'r') as f:
		packages = json.loads(f.read())
		packages[package_name] = {'version': package.config.version, 'source': source}
		if auth:
			packages[package_name]['auth'] = auth

	with open(os.path.realpath(os.path.join('.', 'packages.json')), 'w') as f:
		f.write(json.dumps(packages))

	logger.info('Successfully updated \'packages.json\'.')
	logger.info(
		'Remember to check package docs for any \'vars\' you are required to add to your app.'
	)

	_set_testing(False)


def _packages_rm(*, package_name: str, confirm: bool = True):
	from . import _set_testing

	if confirm:
		confirmation = input(
			f'Are you sure you want to remove package \'{package_name}\'? [yN] '
		)
		if not len(confirmation) or confirmation.lower()[0] != 'y':
			logger.info(f'Cancelled removing package \'{package_name}\'. Exiting.')
			exit(0)

	_set_testing(True)

	try:
		packages_path = os.path.realpath(os.path.join('.', 'packages'))
		package_path = os.path.realpath(os.path.join('.', 'packages', package_name))
		sys.path.insert(0, packages_path)
		package = __import__(package_name)
	except:
		logger.error(
			'Failed to load package to remove it Confirm package name is correct. Exiting.'
		)
		exit(1)

	logger.warning(f'Removing package \'{package_name}\'...')
	# [DOC] Handle removing files for Windows/NT when shutil.rmtree fails
	# [REF] https://stackoverflow.com/a/28476782/2393762
	def errorRemoveReadonly(func, path, exc):
		excvalue = exc[1]
		if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
			os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
			func(path)

	shutil.rmtree(package_path, ignore_errors=False, onerror=errorRemoveReadonly)
	shutil.rmtree(
		os.path.join(packages_path, f'{package_name}-{package.config.version}.dist-info'),
		ignore_errors=False,
		onerror=errorRemoveReadonly,
	)
	logger.info(f'Package \'{package_name}\' removed. Exiting.')
	exit(1)


def packages_audit(args: argparse.Namespace):
	import json

	logger.info(
		'Attempting to audit packages in \'nawah_pacakges.py\' against packages added to \'packages\' folder.'
	)

	app_path = os.path.realpath('.')
	packages_path = os.path.realpath(os.path.join('.', 'packages'))
	sys.path.insert(0, app_path)
	sys.path.append(packages_path)
	import nawah_app

	with open(os.path.join(app_path, 'packages.json')) as f:
		req_packages: Dict[str, Any] = json.loads(f.read())

	logger.info('Packages in \'nawah_pacakges.py\' are:')
	for package_name in req_packages.keys():
		logger.info(f'- {package_name}: {req_packages[package_name]}')

	logger.info('Attempting to check packages added in \'packages\'.')
	added_packages = {}
	for _, pkgname, ispkg in pkgutil.iter_modules([packages_path]):
		if not ispkg:
			continue
		logger.info(f'- Importing package: {pkgname}')
		package = __import__(pkgname, fromlist='*')
		added_packages[pkgname] = package.config.version
		logger.info(f'- Imported package with version: {package.config.version}')

	logger.info('Attempting to audit missing, extra, version-mismatching packages.')

	missing_packages = [
		package for package in req_packages if package not in added_packages.keys()
	]
	if missing_packages:
		logger.info(
			'Following packages are required by your app but not added to \'packages\':'
		)
		for package in missing_packages:
			logger.info(f'- {package}')
	else:
		logger.info('Great! Your app is not missing any of the required package.')

	extra_packages = [
		package
		for package in added_packages
		if package not in req_packages.keys() and package != nawah_app.config.default_package
	]
	if extra_packages:
		logger.info(
			'Following packages are added to \'packages\' but not required by your app:'
		)
		for package in extra_packages:
			logger.info(f'- {package}')
	else:
		logger.info('Great! Your app is not having any extra package.')

	mismatch_packages = [
		package
		for package in req_packages
		if package in added_packages.keys()
		and added_packages[package] != req_packages[package]['version']
	]
	if mismatch_packages:
		logger.warning(
			'Following packages are having mismatch between required and added version:'
		)
		for package in mismatch_packages:
			logger.info(
				f'- {package} requires version \'{req_packages[package]}\' but \'{added_packages[package]}\' is added'
			)
	else:
		logger.info('Great! Your app is not having any version-mismatching package.')



from ._dictobj import DictObj


class BaseModel(DictObj):
	def __repr__(self):
		return f'<Model:{str(self._id)}>'

	def __init__(self, attrs):
		for attr in attrs.keys():
			if type(attrs[attr]) == dict and '_id' in attrs[attr].keys():
				attrs[attr] = BaseModel(attrs[attr])
		super().__init__(attrs)


from nawah.config import Config
from nawah.classes import UnexpectedGatewayException, InvalidGatewayException

import logging, traceback

logger = logging.getLogger('nawah')


class Gateway:
	@staticmethod
	def send(*, gateway: str, **kwargs):
		if Config.test:
			logger.debug('Skipping \'Gateway.send\' action due to test mode.')
			return

		if gateway not in Config.gateways.keys():
			raise InvalidGatewayException(gateway=gateway)

		try:
			Config.gateways[gateway](**kwargs)
		except Exception as e:
			logger.error('Gateway call with following \'kwargs\' failed:')
			logger.error(kwargs)
			logger.error(traceback.format_exc())
			raise UnexpectedGatewayException(gateway=gateway)

		return



from nawah.config import Config
from nawah.classes import NAWAH_ENV, ATTR, NAWAH_DOC

from bson import ObjectId
from typing import Dict, List, Any, Union

import logging, copy

logger = logging.getLogger('nawah')


async def update(
	*,
	env: NAWAH_ENV,
	collection_name: str,
	attrs: Dict[str, ATTR],
	docs: List[Union[str, ObjectId]],
	doc: NAWAH_DOC,
) -> Dict[str, Any]:
	# [DOC] Recreate docs list by converting all docs items to ObjectId
	docs = [ObjectId(doc) for doc in docs]
	# [DOC] Perform update query on matching docs
	collection = env['conn'][Config.data_name][collection_name]
	results = None
	doc = copy.deepcopy(doc)

	# [TODO] Abstract $set pipeline with colon support for all stages

	# [DOC] Prepare empty update pipeline
	update_pipeline: List[Any] = []

	# [DOC] Iterate over attrs in doc to set update stage in pipeline
	for attr in doc.keys():
		# [DOC] Prepare stage pipeline
		update_pipeline_stage_root: Dict[str, Any] = {'$set': {}}
		update_pipeline_stage_current = update_pipeline_stage_root['$set']
		attr_path_part = attr
		attr_path_current = []

		if ':' in attr:
			attr_path = attr.split('.')
			for i in range(len(attr_path)):
				attr_path_part = attr_path[i]
				if i == 0:
					# [DOC] First item has to have $
					attr_path_current.append('$' + attr_path_part.split(':')[0])
				else:
					attr_path_current.append(attr_path_part.split(':')[0])

				if ':' not in attr_path_part:
					part_pipeline: Dict[str, Any] = {
						'$arrayToObject': {
							'$concatArrays': [
								{
									'$objectToArray': '.'.join(attr_path_current[:-1]),
								},
								[
									{
										'k': attr_path_part,
										'v': None,
									}
								],
							]
						}
					}
					if 'v' in update_pipeline_stage_current.keys():
						update_pipeline_stage_current['v'] = part_pipeline
					elif 'then' in update_pipeline_stage_current.keys():
						update_pipeline_stage_current['then'] = part_pipeline
					else:
						update_pipeline_stage_current[attr_path_part] = part_pipeline

					update_pipeline_stage_current = part_pipeline['$arrayToObject']['$concatArrays'][
						1
					][0]
				else:
					part_pipeline = {
						'$map': {
							'input': '.'.join(attr_path_current),
							'as': f'this_{i}',
							'in': {
								'$cond': {
									'if': {
										'$eq': [
											{
												'$indexOfArray': [
													'.'.join(attr_path_current),
													f'$$this_{i}',
												]
											},
											int(attr_path_part.split(':')[1]),
										]
									},
									'then': None,
									'else': f'$$this_{i}',
								}
							},
						}
					}

					if i != 0:
						# [DOC] For all subsequent array objects, wrap in object-to-array-to-object pipeline
						part_pipeline = {
							'$arrayToObject': {
								'$concatArrays': [
									{
										'$objectToArray': '.'.join(attr_path_current[:-1]),
									},
									[
										{
											'k': attr_path_part.split(':')[0],
											'v': part_pipeline,
										}
									],
								]
							}
						}

					if 'v' in update_pipeline_stage_current.keys():
						update_pipeline_stage_current['v'] = part_pipeline
					elif 'then' in update_pipeline_stage_current.keys():
						update_pipeline_stage_current['then'] = part_pipeline
					else:
						update_pipeline_stage_current[attr_path_part.split(':')[0]] = part_pipeline

					if i == 0:
						update_pipeline_stage_current = part_pipeline['$map']['in']['$cond']
					else:
						update_pipeline_stage_current = part_pipeline['$arrayToObject']['$concatArrays'][
							1
						][0]['v']['$map']['in']['$cond']

					attr_path_current = [f'$$this_{i}']

		# [DOC] Check for $add Doc Oper
		if type(doc[attr]) == dict and '$add' in doc[attr].keys():
			add_field = (
				f'${doc[attr]["$field"]}'
				if '$field' in doc[attr].keys() and doc[attr]['$field']
				else f'${".".join(attr_path_current + [attr])}'
			)

			part_pipeline = {
				'$add': [
					{
						'$cond': {
							'if': {'$not': [add_field]},
							'then': 0,
							'else': add_field,
						}
					},
					doc[attr]['$add'],
				]
			}

			# [DOC] Add part_pipeline to update_pipeline_stage_current
			if 'v' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['v'] = part_pipeline
			elif 'then' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['then'] = part_pipeline
			else:
				update_pipeline_stage_current[attr_path_part] = part_pipeline

		# [DOC] Check for $add Doc Oper
		elif type(doc[attr]) == dict and '$multiply' in doc[attr].keys():
			multiply_field = (
				f'${doc[attr]["$field"]}'
				if '$field' in doc[attr].keys() and doc[attr]['$field']
				else f'${".".join(attr_path_current + [attr])}'
			)

			part_pipeline = {
				'$multiply': [
					{
						'$cond': {
							'if': {'$not': [multiply_field]},
							'then': 0,
							'else': multiply_field,
						}
					},
					doc[attr]['$multiply'],
				]
			}

			# [DOC] Add part_pipeline to update_pipeline_stage_current
			if 'v' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['v'] = part_pipeline
			elif 'then' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['then'] = part_pipeline
			else:
				update_pipeline_stage_current[attr_path_part] = part_pipeline

		# [DOC] Check for $append Doc Oper
		elif type(doc[attr]) == dict and '$append' in doc[attr].keys():
			if '$unique' not in doc[attr].keys() or doc[attr]['$unique'] == False:
				part_pipeline = {'$concatArrays': [f'${attr}', [doc[attr]['$append']]]}
			else:
				part_pipeline = {
					'$concatArrays': [
						f'${attr}',
						{
							'$cond': {
								'if': {'$in': [doc[attr]['$append'], f'${attr}']},
								'then': [],
								'else': [doc[attr]['$append']],
							}
						},
					]
				}

			# [DOC] Add part_pipeline to update_pipeline_stage_current
			if 'v' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['v'] = part_pipeline
			elif 'then' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['then'] = part_pipeline
			else:
				update_pipeline_stage_current[attr_path_part] = part_pipeline

		# [DOC] Check for $set_index Doc Oper
		elif type(doc[attr]) == dict and '$set_index' in doc[attr].keys():
			part_pipeline = {
				'$reduce': {
					'input': f'${attr}',
					'initialValue': [],
					'in': {
						'$concatArrays': [
							'$$value',
							{
								'$cond': {
									'if': {
										'$eq': [
											['$$this'],
											[{'$arrayElemAt': [f'${attr}', doc[attr]['$index']]}],
										]
									},
									'then': [doc[attr]['$set_index']],
									'else': ['$$this'],
								}
							},
						]
					},
				}
			}

			# [DOC] Add part_pipeline to update_pipeline_stage_current
			if 'v' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['v'] = part_pipeline
			elif 'then' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['then'] = part_pipeline
			else:
				update_pipeline_stage_current[attr_path_part] = part_pipeline

		# [DOC] Check for $del_val Doc Oper
		elif type(doc[attr]) == dict and '$del_val' in doc[attr].keys():
			part_pipeline = {
				'$reduce': {
					'input': f'${attr}',
					'initialValue': [],
					'in': {
						'$concatArrays': [
							'$$value',
							{
								'$cond': {
									'if': {
										'$eq': [
											['$$this'],
											doc[attr]['$del_val'],
										]
									},
									'then': [],
									'else': ['$$this'],
								}
							},
						]
					},
				}
			}

			# [DOC] Add part_pipeline to update_pipeline_stage_current
			if 'v' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['v'] = part_pipeline
			elif 'then' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['then'] = part_pipeline
			else:
				update_pipeline_stage_current[attr_path_part] = part_pipeline

		# [DOC] Check for $del_index Doc Oper
		elif type(doc[attr]) == dict and '$del_index' in doc[attr].keys():
			part_pipeline = {
				'$arrayToObject': {
					'$reduce': {
						'input': {
							'$objectToArray': f'${attr}',
						},
						'initialValue': [],
						'in': {
							'$concatArrays': [
								'$$value',
								{
									'$cond': {
										'if': {
											'$eq': ['$$this.k', doc[attr]['$del_index']],
										},
										'then': [],
										'else': ['$$this'],
									}
								},
							]
						},
					}
				}
			}

			# [DOC] Add part_pipeline to update_pipeline_stage_current
			if 'v' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['v'] = part_pipeline
			elif 'then' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['then'] = part_pipeline
			else:
				update_pipeline_stage_current[attr_path_part] = part_pipeline

		else:
			# [DOC] Add part_pipeline to update_pipeline_stage_current
			if 'v' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['v'] = {'$literal': doc[attr]}
			elif 'then' in update_pipeline_stage_current.keys():
				update_pipeline_stage_current['then'] = {'$literal': doc[attr]}
			else:
				update_pipeline_stage_current[attr_path_part] = {'$literal': doc[attr]}

		# [DOC] Add stage to pipeline
		update_pipeline.append(update_pipeline_stage_root)

	logger.debug(f'Final update pipeline: {update_pipeline}')

	# [DOC] If using Azure Mongo service update docs one by one
	if Config.data_azure_mongo:
		update_count = 0
		for _id in docs:
			results = await collection.update_one({'_id': _id}, update_pipeline)
			update_count += results.modified_count
	else:
		results = await collection.update_many({'_id': {'$in': docs}}, update_pipeline)
		update_count = results.modified_count

	return {'count': update_count, 'docs': [{'_id': doc} for doc in docs]}



from nawah.base_module import BaseModule
from nawah.enums import Event
from nawah.classes import ATTR, PERM, EXTN, METHOD, InvalidAttrException
from nawah.utils import validate_doc, generate_dynamic_attr
from nawah.config import Config


async def attr_extn_val(
	*,
	mode,
	attr_name,
	attr_type,
	attr_val,
	skip_events,
	env,
	query,
	doc,
	scope,
):
	if type(doc['val']) == dict and '__extn' in doc['val'].keys():
		return {
			'__extn': EXTN(
				module=doc['val']['__extn']['__module'],
				attrs=doc['val']['__extn']['__attrs'],
				force=doc['val']['__extn']['__force'],
			),
			'__val': doc['val']['__extn']['__val'],
		}

	raise InvalidAttrException(
		attr_name=attr_name, attr_type=attr_type, val_type=type(doc['val'])
	)


async def attr_query_mod_type(
	*,
	mode,
	attr_name,
	attr_type,
	attr_val,
	skip_events,
	env,
	query,
	doc,
	scope,
):
	if 'type' not in query or query['type'][0] == 'user_sys':
		raise InvalidAttrException(
			attr_name='type',
			attr_type=ATTR.LITERAL(literal=['global', 'user']),
			val_type=str,
		)


class Setting(BaseModule):
	'''`Setting` module module provides data type and controller for settings in Nawah eco-system. This is used by `User` module tp provide additional user-wise settings. It also allows for global-typed settings.'''

	collection = 'settings'
	attrs = {
		'user': ATTR.ID(desc='`_id` of `User` doc the doc belongs to.'),
		'var': ATTR.STR(
			desc='Name of the setting. This is unique for every `user` in the module.'
		),
		'val': ATTR.ANY(desc='Value of the setting.'),
		'val_type': ATTR.DYNAMIC_ATTR(),
		'type': ATTR.LITERAL(
			desc='Type of the setting. This sets whether setting is global, or belong to user, and whether use can update it or not.',
			literal=['global', 'user', 'user_sys'],
		),
	}
	diff = True
	unique_attrs = [('user', 'var', 'type')]
	extns = {
		'val': ATTR.TYPE(type=attr_extn_val),
	}
	methods = {
		'read': METHOD(
			permissions=[
				PERM(privilege='admin', query_mod={'$limit': 1}),
				PERM(
					privilege='read',
					query_mod={
						'user': '$__user',
						# 'type': ATTR.TYPE(type=attr_query_mod_type),
						'$limit': 1,
					},
				),
			],
			query_args=[
				{
					'_id': ATTR.ID(),
					'type': ATTR.LITERAL(literal=['global', 'user', 'user_sys']),
				},
				{
					'var': ATTR.STR(),
					'type': ATTR.LITERAL(literal=['global']),
				},
				{
					'var': ATTR.STR(),
					'user': ATTR.ID(),
					'type': ATTR.LITERAL(literal=['user', 'user_sys']),
				},
			],
		),
		'create': METHOD(
			permissions=[
				PERM(privilege='admin'),
				PERM(privilege='create', doc_mod={'type': 'user'}),
			]
		),
		'update': METHOD(
			permissions=[
				PERM(privilege='admin', query_mod={'$limit': 1}),
				PERM(
					privilege='update',
					query_mod={'type': 'user', 'user': '$__user', '$limit': 1},
					doc_mod={'var': None, 'val_type': None, 'type': None},
				),
			],
			query_args=[
				{
					'_id': ATTR.ID(),
					'type': ATTR.LITERAL(literal=['global', 'user', 'user_sys']),
				},
				{
					'var': ATTR.STR(),
					'type': ATTR.LITERAL(literal=['global']),
				},
				{
					'var': ATTR.STR(),
					'user': ATTR.ID(),
					'type': ATTR.LITERAL(literal=['user', 'user_sys']),
				},
			],
		),
		'delete': METHOD(
			permissions=[PERM(privilege='admin', query_mod={'$limit': 1})],
			query_args=[{'_id': ATTR.ID()}, {'var': ATTR.STR()}],
		),
		'retrieve_file': METHOD(
			permissions=[PERM(privilege='*', query_mod={'type': 'global'})],
			get_method=True,
		),
	}

	async def on_create(self, results, skip_events, env, query, doc, payload):
		if doc['type'] in ['user', 'user_sys']:
			if doc['user'] == env['session'].user._id and doc['var'] in Config.user_doc_settings:
				env['session'].user[doc['var']] = doc['val']
		return (results, skip_events, env, query, doc, payload)

	async def pre_update(self, skip_events, env, query, doc, payload):
		for attr in doc.keys():
			if attr == 'val' or attr.startswith('val.'):
				val_attr = attr
				break
		else:
			raise self.exception(
				status=400,
				msg='Could not match doc with any of the required doc_args. Failed sets:[\'val\': Missing]',
				args={'code': 'INVALID_DOC'},
			)

		setting_results = await self.read(skip_events=[Event.PERM], env=env, query=query)
		if not setting_results.args.count:
			raise self.exception(
				status=400, msg='Invalid Setting doc', args={'code': 'INVALID_SETTING'}
			)

		setting = setting_results.args.docs[0]
		# [DOC] Attempt to validate val against Setting val_type
		try:
			exception_raised: Exception = None
			setting_val_type, _ = generate_dynamic_attr(dynamic_attr=setting.val_type)
			await validate_doc(
				mode='update',
				doc=doc,
				attrs={'val': setting_val_type},
				skip_events=skip_events,
				env=env,
				query=query,
			)
		except Exception as e:
			exception_raised = e

		if exception_raised or doc[val_attr] == None:
			raise self.exception(
				status=400,
				msg=f'Invalid value for for Setting doc of type \'{type(doc[val_attr])}\' with required type \'{setting.val_type}\'',
				args={'code': 'INVALID_ATTR'},
			)

		return (skip_events, env, query, doc, payload)

	async def on_update(self, results, skip_events, env, query, doc, payload):
		# [TODO] Update according to the changes of Doc Opers
		try:
			if (
				query['type'][0] in ['user', 'user_sys']
				and query['user'][0] == env['session'].user._id
				and query['var'][0] in Config.user_doc_settings
			):
				if type(doc['val']) == dict and '$add' in doc['val'].keys():
					env['session'].user[query['var'][0]] += doc['val']['$add']
				elif type(doc['val']) == dict and '$multiply' in doc['val'].keys():
					env['session'].user[query['var'][0]] *= doc['val']['$multiply']
				elif type(doc['val']) == dict and '$append' in doc['val'].keys():
					env['session'].user[query['var'][0]].append(doc['val']['$append'])
				elif type(doc['val']) == dict and '$set_index' in doc['val'].keys():
					env['session'].user[query['var'][0]][doc['val']['$index']] = doc['val'][
						'$set_index'
					]
				elif type(doc['val']) == dict and '$del_val' in doc['val'].keys():
					for val in doc['val']['$del_val']:
						env['session'].user[query['var'][0]].remove(val)
				elif type(doc['val']) == dict and '$del_index' in doc['val'].keys():
					del env['session'].user[query['var'][0]][doc['val']['$index']]
				else:
					env['session'].user[query['var'][0]] = doc['val']
		except:
			pass
		return (results, skip_events, env, query, doc, payload)



from nawah.enums import LOCALE_STRATEGY

from typing import (
	List,
	Dict,
	Callable,
	Any,
	Union,
	Optional,
	TYPE_CHECKING,
)

if TYPE_CHECKING:
	from nawah.base_module import BaseModule
	from nawah.classes import (
		NAWAH_DOC,
		NAWAH_ENV,
		ATTR,
		CLIENT_APP,
		ANALYTICS_EVENTS,
		SYS_DOC,
		USER_SETTING,
		JOB,
	)

	from motor.motor_asyncio import AsyncIOMotorClient
	from bson import ObjectId
	from datetime import datetime


class Config:
	debug: bool = False
	env: str
	port: int = 8081

	_sys_conn: 'AsyncIOMotorClient'
	_sys_env: 'NAWAH_ENV'
	_sys_docs: Dict['ObjectId', 'SYS_DOC'] = {}
	_jobs_base: 'datetime'

	_nawah_version: str
	packages_api_levels: Dict[str, str] = {}
	packages_versions: Dict[str, str] = {}

	_app_name: str
	_app_version: str
	_app_default_package: str
	_app_path: str
	_app_packages: Dict[str, str]

	test: bool = False

	emulate_test: bool = False
	force_admin_check: bool = False

	generate_ref: bool = False
	_api_ref: str

	generate_models: bool = False
	_api_models: str

	vars_types: Dict[str, Union['ATTR', Dict[str, Any]]] = {}
	vars: Dict[str, Any] = {}

	client_apps: Dict[str, 'CLIENT_APP'] = {}

	analytics_events: 'ANALYTICS_EVENTS' = {
		'app_conn_verified': True,
		'session_conn_auth': True,
		'session_user_auth': True,
		'session_conn_reauth': True,
		'session_user_reauth': True,
		'session_conn_deauth': True,
		'session_user_deauth': True,
	}

	conn_timeout: int = 120
	quota_anon_min: int = 40
	quota_auth_min: int = 100
	quota_ip_min: int = 500
	file_upload_limit: int = -1
	file_upload_timeout: int = 300

	data_server: str = 'mongodb://localhost'
	data_name: str = 'nawah_data'
	data_ssl: bool = False
	data_ca_name: Optional[str] = None
	data_ca: Optional[str] = None
	data_disk_use: bool = False

	data_azure_mongo: bool = False

	locales: List[str] = ['ar_AE', 'en_AE']
	locale: str = 'ar_AE'
	locale_strategy: 'LOCALE_STRATEGY' = LOCALE_STRATEGY.DUPLICATE

	admin_doc: 'NAWAH_DOC' = {}
	admin_password: str = '__0xADMIN'

	anon_token: str = '__ANON_TOKEN_f00000000000000000000012'
	anon_privileges: Dict[str, List[str]] = {}

	user_attrs: Dict[str, 'ATTR'] = {}
	user_settings: Dict[str, 'USER_SETTING'] = {}
	user_doc_settings: List[str] = []

	groups: List[Dict[str, Any]] = []
	default_privileges: Dict[str, List[str]] = {}

	data_indexes: List[Dict[str, Any]] = []

	docs: List['SYS_DOC'] = []

	l10n: Dict[str, Dict[str, Any]] = {}

	jobs: Dict[str, 'JOB'] = {}

	gateways: Dict[str, Callable] = {}

	types: Dict[str, Callable] = {}

	modules: Dict[str, 'BaseModule'] = {}
	modules_packages: Dict[str, List[str]] = {}



import inspect
import logging
import os.path
import pkgutil
import re
import sys

import nawah
from nawah.classes import ATTR, L10N
from nawah.config import Config

from ._config import _process_config
from ._validate import validate_attr

logger = logging.getLogger('nawah')


async def _import_modules():
    sys.path.append(os.path.join(nawah.__path__[0], 'packages'))
    sys.path.append(os.path.join(Config._app_path, 'packages'))

    # [DOC] Iterate over packages in modules folder
    for _, pkgname, ispkg in pkgutil.iter_modules(
        [
            os.path.join(nawah.__path__[0], 'packages'),
            os.path.join(Config._app_path, 'packages'),
        ]
    ):
        if not ispkg:
            continue

        logger.debug(f'Importing package: {pkgname}')

        # [DOC] Load package and attempt to load config
        package = __import__(pkgname, fromlist='*')
        _process_config(config=package.config, pkgname=pkgname)

        # [DOC] Add package to loaded packages dict
        Config.modules_packages[pkgname] = []

        # [DOC] Iterate over python modules in package
        package_prefix = package.__name__ + '.'
        for _, modname, ispkg in pkgutil.iter_modules(package.__path__, package_prefix):
            # [DOC] Iterate over Python modules in package
            module = __import__(modname, fromlist='*')
            # [DOC] For __l10n__ Python module, extract all l10n dicts and skip processing
            if modname.endswith('__l10n__'):
                for l10n_name in dir(module):
                    if type(getattr(module, l10n_name)) == L10N:
                        if l10n_name not in Config.l10n.keys():
                            Config.l10n[l10n_name] = {}
                        Config.l10n[l10n_name].update(getattr(module, l10n_name))
                continue

            for clsname in dir(module):
                # [DOC] Confirm class is subclass of BaseModule
                if (
                    clsname != 'BaseModule'
                    and inspect.isclass(getattr(module, clsname))
                    and getattr(getattr(module, clsname), '_nawah_module', False)
                ):
                    # [DOC] Deny loading Nawah-reserved named Nawah modules
                    if clsname.lower() in ['conn', 'heart', 'watch']:
                        logger.error(
                            f'Module with Nawah-reserved name \'{clsname.lower()}\' was found. Exiting.'
                        )
                        exit(1)
                    # [DOC] Load Nawah module and assign module_name attr
                    cls = getattr(module, clsname)
                    module_name = re.sub(
                        r'([A-Z])', r'_\1', clsname[0].lower() + clsname[1:]
                    ).lower()
                    # [DOC] Deny duplicate Nawah modules names
                    if module_name in Config.modules.keys():
                        logger.error(
                            f'Duplicate module name \'{module_name}\'. Exiting.'
                        )
                        exit(1)
                    # [DOC] Add module to loaded modules dict
                    Config.modules[module_name] = cls()
                    Config.modules_packages[pkgname].append(module_name)

    # [DOC] Update User, Session modules with populated attrs
    Config.modules['user'].attrs.update(Config.user_attrs)
    if (
        sum(
            1
            for attr in Config.user_settings.keys()
            if attr in Config.user_attrs.keys()
        )
        != 0
    ):
        logger.error(
            'At least on attr from \'user_settings\' is conflicting with an attr from \'user_attrs\'. Exiting.'
        )
        exit(1)
    for attr in Config.user_doc_settings:
        Config.modules['user'].attrs[attr] = Config.user_settings[attr].val_type
        Config.modules['user'].attrs[attr]._default = Config.user_settings[attr].default
    Config.modules['user'].defaults['locale'] = Config.locale
    for attr in Config.user_attrs.keys():
        Config.modules['user'].unique_attrs.append(attr)
        Config.modules['user'].attrs[f'{attr}_hash'] = ATTR.STR()
        Config.modules['session'].methods['auth'].doc_args.append(
            {
                'hash': ATTR.STR(),
                attr: Config.user_attrs[attr],
                'groups': ATTR.LIST(list=[ATTR.ID()]),
            }
        )
        Config.modules['session'].methods['auth'].doc_args.append(
            {'hash': ATTR.STR(), attr: Config.user_attrs[attr]}
        )

    # [DOC] Attempt to validate all packages required vars (via vars_types Config Attr) are met
    for var in Config.vars_types.keys():
        if var not in Config.vars.keys():
            logger.error(
                f'Package \'{Config.vars_types[var]["package"]}\' requires \'{var}\' Var, but not found in App Config. Exiting.'
            )
            exit(1)
        try:
            await validate_attr(
                mode='create',
                attr_name=var,
                attr_type=Config.vars_types[var]['type'],
                attr_val=Config.vars[var],
            )
        except:
            logger.error(
                f'Package \'{Config.vars_types[var]["package"]}\' requires \'{var}\' Var of type \'{Config.vars_types[var]["type"]._type}\', but validation failed. Exiting.'
            )
            exit(1)

    # [DOC] Call update_modules, effectively finalise initialising modules
    for module in Config.modules.values():
        module._initialise()



from nawah.enums import Event

from asyncio import Task
from motor.motor_asyncio import AsyncIOMotorClient
from aiohttp.web import WebSocketResponse
from typing import (
	List,
	TypedDict,
	Optional,
	Dict,
	Callable,
	Literal,
	Union,
	Any,
	AsyncGenerator,
	TYPE_CHECKING,
	Coroutine,
)

import datetime

if TYPE_CHECKING:
	from ._base_model import BaseModel
	from ._dictobj import DictObj

NAWAH_EVENTS = List[Event]

WATCH_TASK = TypedDict(
	'WATCH_TASK',
	{
		'watch': Coroutine[Any, Any, None],
		'task': Task,
		'stream': Any,
	},
	total=False,
)

NAWAH_ENV_QUOTA = TypedDict(
	'NAWAH_ENV_QUOTA',
	{
		'counter': int,
		'last_check': datetime.datetime,
	},
)

NAWAH_ENV = TypedDict(
	'NAWAH_ENV',
	{
		'id': int,
		'init': bool,
		'conn': AsyncIOMotorClient,
		'REMOTE_ADDR': str,
		'HTTP_USER_AGENT': str,
		'HTTP_ORIGIN': str,
		'client_app': str,
		'session': 'BaseModel',
		'last_call': datetime.datetime,
		'ws': WebSocketResponse,
		'watch_tasks': Dict[str, WATCH_TASK],
		'quota': NAWAH_ENV_QUOTA,
	},
	total=False,
)

NAWAH_QUERY_SPECIAL_GROUP = TypedDict(
	'NAWAH_QUERY_SPECIAL_GROUP', {'by': str, 'count': int}
)

NAWAH_QUERY_SPECIAL_GEO_NEAR = TypedDict(
	'NAWAH_QUERY_SPECIAL_GEO_NEAR', {'val': str, 'attr': str, 'dist': int}
)

NAWAH_QUERY_SPECIAL = TypedDict(
	'NAWAH_QUERY_SPECIAL',
	{
		'$search': Optional[str],
		'$sort': Optional[Dict[str, Literal[1, -1]]],
		'$skip': Optional[int],
		'$limit': Optional[int],
		'$extn': Optional[Union[Literal[False], List[str]]],
		'$attrs': Optional[List[str]],
		'$group': Optional[List[NAWAH_QUERY_SPECIAL_GEO_NEAR]],
		'$geo_near': Optional[NAWAH_QUERY_SPECIAL_GEO_NEAR],
	},
	total=False,
)

NAWAH_QUERY = List[  # type: ignore
	Union[
		'NAWAH_QUERY',  # type: ignore
		Union[
			Dict[
				str,
				Union[
					'NAWAH_QUERY',  # type: ignore
					Any,
					Union[
						Dict[Literal['$eq'], Any],
						Dict[Literal['$ne'], Any],
						Dict[Literal['$gt'], Union[int, str]],
						Dict[Literal['$gte'], Union[int, str]],
						Dict[Literal['$lt'], Union[int, str]],
						Dict[Literal['$lte'], Union[int, str]],
						Dict[Literal['$bet'], Union[List[int], List[str]]],
						Dict[Literal['$all'], List[Any]],
						Dict[Literal['$in'], List[Any]],
						Dict[Literal['$nin'], List[Any]],
						Dict[Literal['$regex'], str],
					],
				],
			],
			NAWAH_QUERY_SPECIAL,
		],
	]
]

NAWAH_DOC = Dict[
	str,
	Union[
		Dict[
			str,
			Union[
				Dict[Literal['$add'], int],
				Dict[Literal['$multiply'], int],
				Dict[Literal['$append'], Any],
				Dict[Literal['$set_index'], Dict[int, Any]],
				Dict[Literal['$set_index'], Dict[int, Any]],
				Dict[Literal['$del_val'], Any],
				Dict[Literal['$del_index'], int],
				Any,
			],
		],
		Any,
	],
]

IP_QUOTA = TypedDict(
	'IP_QUOTA',
	{
		'counter': int,
		'last_check': datetime.datetime,
	},
)


from nawah.config import Config
from nawah.enums import LOCALE_STRATEGY, Event
from nawah.utils import _extract_attr
from nawah.classes import (
	NAWAH_ENV,
	ATTR,
	Query,
	BaseModel,
	NAWAH_DOC,
	EXTN,
	InvalidAttrException,
)
from ._query import _compile_query

from motor.motor_asyncio import AsyncIOMotorCollection
from bson import ObjectId
from typing import Dict, Any, Union, List, Optional, cast

import logging, copy

logger = logging.getLogger('nawah')


async def read(
	*,
	env: NAWAH_ENV,
	collection_name: str,
	attrs: Dict[str, ATTR],
	query: Query,
	skip_process: bool = False,
	skip_extn: bool = False,
) -> Dict[str, Any]:
	skip, limit, sort, group, aggregate_query = _compile_query(
		collection_name=collection_name, attrs=attrs, query=query, watch_mode=False
	)

	logger.debug(f'aggregate_query: {aggregate_query}')
	logger.debug(f'skip, limit, sort, group: {skip}, {limit}, {sort}, {group}.')

	collection: AsyncIOMotorCollection = env['conn'][Config.data_name][collection_name]
	docs_total_results = collection.aggregate(
		aggregate_query + [{'$count': '__docs_total'}],
		allowDiskUse=Config.data_disk_use,
	)
	try:
		async for doc in docs_total_results:
			docs_total = doc['__docs_total']
		docs_total
	except:
		return {'total': 0, 'count': 0, 'docs': [], 'groups': []}

	groups = {}
	if group:
		for group_condition in group:
			group_query = aggregate_query + [
				{
					'$bucketAuto': {
						'groupBy': '$' + group_condition['by'],
						'buckets': group_condition['count'],
					}
				}
			]
			check_group = False
			for i in range(len(group_query)):
				if (
					list(group_query[i].keys())[0] == '$match'
					and list(group_query[i]['$match'].keys())[0] == group_condition['by']
				):
					check_group = True
					break
			if check_group:
				del group_query[i]
			group_query_results = collection.aggregate(
				group_query, allowDiskUse=Config.data_disk_use
			)
			groups[group_condition['by']] = [
				{
					'min': group['_id']['min'],
					'max': group['_id']['max'],
					'count': group['count'],
				}
				async for group in group_query_results
			]

	if sort != None:
		aggregate_query.append({'$sort': sort})
	if skip != None:
		aggregate_query.append({'$skip': skip})
	if limit != None:
		aggregate_query.append({'$limit': limit})

	logger.debug(f'final query: {collection}, {aggregate_query}.')

	docs_count_results = collection.aggregate(
		aggregate_query + [{'$count': '__docs_count'}],
		allowDiskUse=Config.data_disk_use,
	)
	try:
		async for doc in docs_count_results:
			docs_count = doc['__docs_count']
		docs_count
	except:
		return {
			'total': docs_total,
			'count': 0,
			'docs': [],
			'groups': {} if not group else groups,
		}
	docs = collection.aggregate(aggregate_query, allowDiskUse=Config.data_disk_use)
	models = []
	extn_models: Dict[str, Optional[BaseModel]] = {}
	async for doc in docs:
		if not skip_process:
			doc = await _process_results_doc(
				env=env,
				collection=collection,
				attrs=attrs,
				doc=doc,
				extn_models=extn_models,
				skip_extn=skip_extn,
			)
		if doc:
			models.append(BaseModel(doc))
	return {
		'total': docs_total,
		'count': docs_count,
		'docs': models,
		'groups': {} if not group else groups,
	}


async def _process_results_doc(
	*,
	env: NAWAH_ENV,
	collection: str,
	attrs: Dict[str, ATTR],
	doc: NAWAH_DOC,
	extn_models: Dict[str, Optional[BaseModel]] = {},
	skip_extn: bool = False,
) -> Dict[str, Any]:
	# [DOC] Process doc attrs
	for attr in attrs.keys():
		if attrs[attr]._type == 'LOCALE':
			if (
				attr in doc.keys() and type(doc[attr]) == dict and Config.locale in doc[attr].keys()
			):
				if Config.locale_strategy == LOCALE_STRATEGY.NONE_VALUE:
					doc[attr] = {
						locale: doc[attr][locale] if locale in doc[attr].keys() else None
						for locale in Config.locales
					}
				elif callable(Config.locale_strategy):
					doc[attr] = {
						locale: doc[attr][locale]
						if locale in doc[attr].keys()
						else Config.locale_strategy(attr_val=doc[attr][Config.locale], locale=locale)
						for locale in Config.locales
					}
				else:
					doc[attr] = {
						locale: doc[attr][locale]
						if locale in doc[attr].keys()
						else doc[attr][Config.locale]
						for locale in Config.locales
					}
		if not skip_extn:
			await _extend_attr(
				doc=doc, scope=doc, attr_name=attr, attr_type=attrs[attr], env=env
			)
	# [DOC] Attempt to extned the doc per extns
	return doc


async def _extend_attr(
	*,
	doc: NAWAH_DOC,
	scope: Union[Dict[str, Any], List[Any]],
	attr_name: Union[str, int],
	attr_type: ATTR,
	env: NAWAH_ENV,
	extn_models: Dict[str, Optional[BaseModel]] = None,
):
	if not extn_models:
		extn_models = {}

	# [DOC] If scope is missing attr_name skip
	if type(scope) == dict and attr_name not in scope.keys():  # type: ignore
		return

	# [DOC] Check attr_type for possible types that require deep checking for extending
	if attr_type._type == 'KV_DICT':
		attr_name = cast(str, attr_name)
		scope = cast(Dict[str, Any], scope)
		if scope[attr_name] and type(scope[attr_name]) == dict:
			for child_attr in scope[attr_name].keys():
				# [DOC] attr_type is KV_DICT where Attr Type Arg val could be extended
				await _extend_attr(
					doc=doc,
					scope=scope[attr_name],
					attr_name=child_attr,
					attr_type=attr_type._args['val'],
					env=env,
					extn_models=extn_models,
				)
	if attr_type._type == 'TYPED_DICT':
		attr_name = cast(str, attr_name)
		scope = cast(Dict[str, Any], scope)
		if scope[attr_name] and type(scope[attr_name]) == dict:
			for child_attr in attr_type._args['dict'].keys():
				# [DOC] attr_type is TYPED_DICT where each dict item could be extended
				await _extend_attr(
					doc=doc,
					scope=scope[attr_name],
					attr_name=child_attr,
					attr_type=attr_type._args['dict'][child_attr],
					env=env,
					extn_models=extn_models,
				)

	elif attr_type._type == 'LIST':
		attr_name = cast(str, attr_name)
		scope = cast(Dict[str, Any], scope)
		if scope[attr_name] and type(scope[attr_name]) == list:
			for child_attr in attr_type._args['list']:
				# [DOC] attr_type is LIST where it could have KV_DICT, TYPED_DICT, ID Attrs Types that can be [deep-]extended
				if child_attr._type == 'KV_DICT':
					for child_scope in scope[attr_name]:
						if type(child_scope) == dict:
							for child_child_attr in child_scope.keys():
								await _extend_attr(
									doc=doc,
									scope=child_scope,
									attr_name=child_child_attr,
									attr_type=child_attr._args['val'],
									env=env,
									extn_models=extn_models,
								)
				elif child_attr._type == 'TYPED_DICT':
					for child_scope in scope[attr_name]:
						if type(child_scope) == dict:
							for child_child_attr in child_attr._args['dict'].keys():
								await _extend_attr(
									doc=doc,
									scope=child_scope,
									attr_name=child_child_attr,
									attr_type=child_attr._args['dict'][child_child_attr],
									env=env,
									extn_models=extn_models,
								)
				elif child_attr._type == 'ID':
					for i in range(len(scope[attr_name])):
						await _extend_attr(
							doc=doc,
							scope=scope[attr_name],
							attr_name=i,
							attr_type=child_attr,
							env=env,
							extn_models=extn_models,
						)

	# [DOC] Attempt to extend the attr unto doc
	if type(attr_type._extn) == ATTR:
		attr_name = cast(str, attr_name)
		scope = cast(Dict[str, Any], scope)
		attr_type._extn = cast(ATTR, attr_type._extn)
		# [DOC] Attr is having Attr Type TYPE (checked by BaseModule.initialise) for _extn value, resolve using the callable
		try:
			extn_set = await attr_type._extn._args['func'](
				mode='create',
				attr_name=attr_name,
				attr_type=attr_type._extn,
				attr_val=scope[attr_name],
				skip_events=[],
				env=env,
				query=[],
				doc=doc,
				scope=scope,
			)

			if type(extn_set['__val']) == ObjectId:
				scope[attr_name] = await _extend_doc(
					env=env,
					doc=doc,
					attr=scope[attr_name],
					extn_id=extn_set['__val'],
					extn=extn_set['__extn'],
					extn_models=extn_models,
				)
			elif type(extn_set['__val']) == list:
				scope[attr_name] = [
					await _extend_doc(
						env=env,
						doc=doc,
						attr=scope[attr_name],
						extn_id=extn_id,
						extn=extn_set['__extn'],
						extn_models=extn_models,
					)
					for extn_id in extn_set['__val']
				]
		except InvalidAttrException as e:
			logger.debug(
				f'Skipping extending attr \'{attr_name}\' due to \'InvalidAttrException\' by Attr Type TYPE.'
			)

	elif type(attr_type._extn) == EXTN:
		attr_name = cast(str, attr_name)
		scope = cast(Dict[str, Any], scope)
		attr_type._extn = cast(EXTN, attr_type._extn)
		# [DOC] Attr is having EXTN for _extn value, attempt to extend attr based on scope type
		if type(scope[attr_name]) == ObjectId:
			scope[attr_name] = await _extend_doc(
				env=env,
				doc=doc,
				attr=scope[attr_name],
				extn_id=scope[attr_name],
				extn=attr_type._extn,
				extn_models=extn_models,
			)
		elif type(scope[attr_name]) == list:
			scope[attr_name] = [
				await _extend_doc(
					env=env,
					doc=doc,
					attr=scope[attr_name],
					extn_id=extn_id,
					extn=attr_type._extn,
					extn_models=extn_models,
				)
				for extn_id in scope[attr_name]
			]


async def _extend_doc(
	*,
	env: NAWAH_ENV,
	doc: NAWAH_DOC,
	attr: Optional[NAWAH_DOC],
	extn_id: ObjectId,
	extn: EXTN,
	extn_models: Dict[str, Optional[BaseModel]] = {},
) -> Optional[BaseModel]:
	# [DOC] Check if extn module is dynamic value
	if extn.module.startswith('$__'):
		extn_module = Config.modules[
			_extract_attr(scope={'doc': doc, 'attr': attr}, attr_path=extn.module)
		]
	else:
		extn_module = Config.modules[extn.module]
	# [DOC] Check if extn attr set to fetch all or specific attrs
	if type(extn.attrs) == str and extn.attrs.startswith('$__'):  # type: ignore
		extn.attrs = cast(str, extn.attrs)
		extn_attrs = _extract_attr(scope={'doc': doc, 'attr': attr}, attr_path=extn.attrs)
		if extn_attrs[0] == '*':
			extn_attrs = {attr: extn_module.attrs[attr] for attr in extn_module.attrs.keys()}
	elif extn.attrs[0] == '*':
		extn_attrs = {attr: extn_module.attrs[attr] for attr in extn_module.attrs.keys()}
	else:
		extn_attrs = {attr: extn_module.attrs[attr] for attr in extn.attrs}
	# [DOC] Implicitly add _id key to extn attrs so that we don't delete it in process
	extn_attrs['_id'] = 'id'
	# [DOC] Set skip events
	skip_events = [Event.PERM]
	# [DOC] Check if extn instruction is explicitly requires second-dimension extn.
	if extn.force == False:
		skip_events.append(Event.EXTN)
	elif type(extn.force) == str and extn.force.startswith('$__'):  # type: ignore
		extn.force = cast(str, extn.force)
		if not _extract_attr(scope={'doc': doc, 'attr': attr}, attr_path=extn.force):
			skip_events.append(Event.EXTN)
	# [DOC] Read doc if not in extn_models
	if str(extn_id) not in extn_models.keys():
		extn_results = await extn_module.methods['read'](
			skip_events=skip_events + (extn.skip_events or []),
			env=env,
			query=[{'_id': extn_id}] + (extn.query or []),  # type: ignore
		)
		if extn_results['args']['count']:
			extn_models[str(extn_id)] = extn_results['args']['docs'][0]
		else:
			extn_models[str(extn_id)] = None
	# [DOC] Set attr to extn_models doc
	extn_doc = copy.deepcopy(extn_models[str(extn_id)])
	# [DOC] delete all unneeded keys from the resulted doc
	if extn_doc:
		extn_doc = BaseModel(
			{attr: extn_doc[attr] for attr in extn_attrs.keys() if attr in extn_doc}
		)
	return extn_doc



from nawah.config import Config
from nawah.utils import _extract_attr
from nawah.classes import (
	SYS_DOC,
	JOB,
	InvalidModuleException,
	InvalidLocaleException,
	InvalidLocaleTermException,
	InvalidVarException,
)

from typing import Dict, List, Any, TYPE_CHECKING

import logging

if TYPE_CHECKING:
	from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


class Registry:
	docs: List[SYS_DOC] = Config.docs
	jobs: Dict[str, JOB] = Config.jobs

	@staticmethod
	def module(module: str) -> 'BaseModule':
		try:
			return Config.modules[module]
		except KeyError:
			raise InvalidModuleException(module=module)

	@staticmethod
	def l10n(*, locale: str, term: str):
		if locale not in Config.l10n.keys():
			raise InvalidLocaleException(locale=locale)

		try:
			if '.' in term:
				return _extract_attr(scope=Config.l10n[locale], attr_path=term)
			else:
				return Config.l10n[locale][term]
		except (KeyError, TypeError):
			raise InvalidLocaleTermException(locale=locale, term=term)

	@staticmethod
	def var(var: str):
		try:
			if '.' in var:
				return _extract_attr(scope=Config.vars, attr_path=var)
			else:
				return Config.vars[var]
		except (KeyError, TypeError):
			raise InvalidVarException(var=var)



from nawah.base_module import BaseModule
from nawah.classes import ATTR, PERM, METHOD
from nawah.config import Config
from nawah.registry import Registry
from nawah.utils import _extract_lambda_body


class Core(BaseModule):
	'''`Core` module provides access to ADMIN user to fetch `Nawah` instance.'''

	methods = {
		'retrieve_config': METHOD(
			permissions=[PERM(privilege='admin')],
			query_args={'config_attr': ATTR.STR()},
		),
		'retrieve_cache_sets': METHOD(
			permissions=[PERM(privilege='admin')],
			query_args={'module': ATTR.STR()},
		),
		'retrieve_cache_queries': METHOD(
			permissions=[PERM(privilege='admin')],
			query_args={'module': ATTR.STR(), 'cache_set': ATTR.INT()},
		),
		'retrieve_cache_results': METHOD(
			permissions=[PERM(privilege='admin')],
			query_args={
				'module': ATTR.STR(),
				'cache_set': ATTR.INT(),
				'query': ATTR.INT(),
			},
		),
	}

	async def retrieve_config(self, skip_events=[], env={}, query=[], doc={}):
		return self.status(
			status=200,
			msg='Config Attr value retrieved.',
			args={'value': getattr(Config, query['config_attr'][0])},
		)

	async def retrieve_cache_sets(self, skip_events=[], env={}, query=[], doc={}):
		return self.status(
			status=200,
			msg='Module Cache Sets retrieved.',
			args={
				'sets': [
					_extract_lambda_body(cache_set.condition)
					for cache_set in Registry.module(query['module'][0]).cache
				]
			},
		)

	async def retrieve_cache_queries(self, skip_events=[], env={}, query=[], doc={}):
		return self.status(
			status=200,
			msg='Module Cache Sets queries retrieved.',
			args={
				'queries': list(
					Registry.module(query['module'][0]).cache[query['cache_set'][0]].queries.keys()
				)
			},
		)

	async def retrieve_cache_results(self, skip_events=[], env={}, query=[], doc={}):
		cache_set_query = list(
			Registry.module(query['module'][0]).cache[query['cache_set'][0]].queries.keys()
		)[query['query'][0]]
		return self.status(
			status=200,
			msg='Module Cache Sets results retrieved.',
			args={
				'results': Registry.module(query['module'][0])
				.cache[query['cache_set'][0]]
				.queries[cache_set_query]
				.results,
				'query_time': Registry.module(query['module'][0])
				.cache[query['cache_set'][0]]
				.queries[cache_set_query]
				.query_time.isoformat(),
			},
		)



from nawah.base_module import BaseModule
from nawah.enums import Event
from nawah.classes import ATTR, PERM, METHOD
from nawah.config import Config
from nawah.registry import Registry
from nawah.utils import _extract_attr, validate_attr

from bson import ObjectId

import base64


class File(BaseModule):
	'''`File` module provides functionality for `File Upload Workflow`.'''

	collection = 'files'
	attrs = {
		'user': ATTR.ID(desc='`_id` of `User` doc file belongs to.'),
		'file': ATTR.FILE(desc='File object.'),
		'create_time': ATTR.DATETIME(
			desc='Python `datetime` ISO format of the doc creation.'
		),
	}
	methods = {
		'read': METHOD(permissions=[PERM(privilege='__sys')]),
		'create': METHOD(
			permissions=[PERM(privilege='create')],
			post_method=True,
		),
		'delete': METHOD(permissions=[PERM(privilege='__sys')]),
	}

	async def on_read(self, results, skip_events, env, query, doc, payload):
		for i in range(len(results['docs'])):
			results['docs'][i]['file']['lastModified'] = int(
				results['docs'][i]['file']['lastModified']
			)
		return (results, skip_events, env, query, doc, payload)

	async def pre_create(self, skip_events, env, query, doc, payload):
		if Config.file_upload_limit != -1 and len(doc['file']) > Config.file_upload_limit:
			raise self.exception(
				status=400,
				msg=f'File size is beyond allowed limit.',
				args={
					'code': 'INVALID_SIZE',
					'attr': doc['__attr'].decode('utf-8'),
					'name': doc['name'].decode('utf-8'),
				},
			)
		if (module := doc['__module'].decode('utf-8')) not in Config.modules.keys():
			raise self.exception(
				status=400,
				msg=f'Invalid module \'{module}\'',
				args={'code': 'INVALID_MODULE'},
			)

		try:
			attr_type = _extract_attr(
				scope=Registry.module(module).attrs,
				attr_path='$__' + (attr := doc['__attr'].decode('utf-8')),
			)
			doc = {
				'file': {
					'name': doc['name'].decode('utf-8'),
					'type': doc['type'].decode('utf-8'),
					'size': len(doc['file']),
					'lastModified': int(doc['lastModified'].decode('utf-8')),
					'content': doc['file'],
				},
			}
			try:
				attr_val = doc['file']
				if attr_type._type == 'LIST':
					attr_val = [doc['file']]
				await validate_attr(
					mode='create', attr_name=attr, attr_type=attr_type, attr_val=attr_val
				)
			except:
				raise self.exception(
					status=400,
					msg=f'Invalid file for \'{attr}\' of module \'{module}\'',
					args={'code': 'INVALID_FILE'},
				)

		except:
			raise self.exception(
				status=400,
				msg=f'Invalid attr \'{attr}\' of module \'{module}\'',
				args={'code': 'INVALID_ATTR'},
			)

		return (skip_events, env, query, doc, payload)



from nawah import __version__

from typing import Dict, Literal, Any, Optional

import argparse, os, logging, datetime, sys, subprocess, asyncio, traceback, shutil, urllib.request, re, tarfile, string, random, tempfile, pkgutil, glob


def nawah_cli():
	from ._logging import logger
	from ._create import create
	from ._launch import launch
	from ._packages import packages_audit, packages_install, _packages_add, _packages_rm
	from ._generate import generate_ref, generate_models

	if sys.version_info.major != 3 or sys.version_info.minor < 8:
		print('Nawah framework CLI can only run with Python >= 3.8. Exiting.')
		exit(1)

	parser = argparse.ArgumentParser()
	parser.add_argument(
		'--version',
		help='Show Nawah framework version and exit',
		action='version',
		version=f'Nawah framework v{__version__}',
	)

	subparsers = parser.add_subparsers(
		title='Command', description='Nawah framework CLI command to run', dest='command'
	)

	parser_create = subparsers.add_parser('create', help='Create new Nawah app')
	parser_create.set_defaults(func=create)
	parser_create.add_argument('app_name', type=str, help='Name of the app to create')
	parser_create.add_argument(
		'app_path',
		type=str,
		nargs='?',
		help='Path to create new Nawah app. [default .]',
		default='.',
	)
	parser_create.add_argument(
		'--default-config',
		help='Create new Nawah app with default config',
		action='store_true',
	)

	parser_launch = subparsers.add_parser('launch', help='Launch Nawah app')
	parser_launch.set_defaults(func=launch)
	parser_launch.add_argument('--env', help='Choose specific env')
	parser_launch.add_argument('--debug', help='Enable debug mode', action='store_true')
	parser_launch.add_argument(
		'--log',
		help='Enable debug mode and log all debug messages to log file',
		action='store_true',
	)
	parser_launch.add_argument('-p', '--port', help='Set custom port [default 8081]')
	parser_launch.add_argument(
		'--force-admin-check',
		help='Force ADMIN doc checked and updated, if ADMIN doc is changed',
		action='store_true',
	)

	parser_packages = subparsers.add_parser('packages', help='Manage Nawah app packages')
	parser_packages.set_defaults(func=lambda _: None)
	packages_subparser = parser_packages.add_subparsers(
		title='Packages Command',
		description='Packages command to run',
		dest='packages_command',
	)

	parser_packages_install = packages_subparser.add_parser(
		'install', help='Install Nawah app packages'
	)
	parser_packages_install.set_defaults(func=packages_install)

	parser_packages_add = packages_subparser.add_parser(
		'add', help='Add package to Nawah app'
	)
	parser_packages_add.set_defaults(
		func=lambda args: _packages_add(
			package_name=args.package_name,
			source=args.source,
			version=args.version,
			auth=args.auth,
		)
	)
	parser_packages_add.add_argument(
		'package_name', help='Package name to add to Nawah app'
	)
	parser_packages_add.add_argument(
		'--source',
		help='Package source [default https://gitlab.com/api/v4/projects/24381550/packages/pypi/simple]',
		default='https://gitlab.com/api/v4/projects/24381550/packages/pypi/simple',
	)
	parser_packages_add.add_argument(
		'--version',
		help='Package version (repo tag name) to install [default latest]',
		default='latest',
	)
	parser_packages_add.add_argument(
		'--auth',
		help='String representing colon-separated username and password combination to authorise the source',
	)

	parser_packages_rm = packages_subparser.add_parser(
		'rm', help='Remove package from Nawah app'
	)
	parser_packages_rm.set_defaults(
		func=lambda args: _packages_rm(package_name=args.package_name, confirm=not args.y)
	)
	parser_packages_rm.add_argument(
		'package_name', help='Package name to remove from Nawah app'
	)
	parser_packages_rm.add_argument('-y', help='Skip confirmation', action='store_true')

	parser_packages_audit = packages_subparser.add_parser(
		'audit', help='Audit packages in Nawah app'
	)
	parser_packages_audit.set_defaults(func=packages_audit)

	parser_ref = subparsers.add_parser('generate_ref', help='Generate Nawah app reference')
	parser_ref.set_defaults(func=generate_ref)
	parser_ref.add_argument('--debug', help='Enable debug mode', action='store_true')

	parser_ref = subparsers.add_parser('generate_models', help='Generate Nawah app models')
	parser_ref.set_defaults(func=generate_models)
	parser_ref.add_argument('format', help='Format of models', choices=['js', 'ts'])
	parser_ref.add_argument('--debug', help='Enable debug mode', action='store_true')

	args = parser.parse_args()

	if args.command:
		if args.command == 'packages' and not args.packages_command:
			parser_packages.print_help()
		else:
			args.func(args)
	else:
		parser.print_help()



from nawah.base_module import BaseModule
from nawah.enums import Event
from nawah.classes import ATTR, PERM, NAWAH_DOC, METHOD
from nawah.registry import Registry

from bson import ObjectId


class Diff(BaseModule):
	'''`Diff` module provides data type and controller for `Diff Workflow`. It is meant for use by internal calls only.'''

	collection = 'diff'
	attrs = {
		'user': ATTR.ID(desc='`_id` of `User` doc the doc belongs to.'),
		'module': ATTR.STR(desc='Name of the module the original doc is part of.'),
		'doc': ATTR.ID(desc='`_id` of the original doc.'),
		'vars': ATTR.KV_DICT(
			desc='Key-value `dict` containing all attrs that have been updated from the original doc.',
			key=ATTR.STR(),
			val=ATTR.ANY(),
		),
		'remarks': ATTR.STR(
			desc='Human-readable remarks of the doc. This is introduced to allow developers to add log messages to diff docs.'
		),
		'create_time': ATTR.DATETIME(
			desc='Python `datetime` ISO format of the doc creation.'
		),
	}
	defaults = {'doc': None, 'remarks': ''}
	methods = {
		'read': METHOD(permissions=[PERM(privilege='read')]),
		'create': METHOD(permissions=[PERM(privilege='__sys')]),
		'delete': METHOD(permissions=[PERM(privilege='delete')]),
	}

	async def pre_create(self, skip_events, env, query, doc, payload):
		# [DOC] format Doc Oper with prefixed underscores to avoid data errors
		doc = self.format_doc_oper(doc=doc)
		# [DOC] Detect non-_id update query:
		if '_id' not in query:
			results = await Registry.module(doc['module']).read(
				skip_events=[Event.PERM], env=env, query=query
			)
			if results.args.count > 1:
				query.append({'_id': {'$in': [doc._id for doc in results.args.docs]}})
			elif results.args.count == 1:
				query.append({'_id': results.args.docs[0]._id})
			else:
				raise self.exception(
					status=400, msg='No update docs matched.', args={'code': 'NO_MATCH'}
				)
		if '_id' in query and type(query['_id'][0]) == list:
			for i in range(len(query['_id'][0]) - 1):
				self.create(
					skip_events=[Event.PERM],
					env=env,
					query=[{'_id': query['_id'][0][i]}],
					doc=doc,
				)
			query['_id'][0] = query['_id'][0][-1]
		doc['doc'] = ObjectId(query['_id'][0])
		return (skip_events, env, query, doc, payload)

	def format_doc_oper(self, *, doc: NAWAH_DOC):
		shadow_doc: NAWAH_DOC = {}
		for attr in doc.keys():
			if attr[0] == '$':
				shadow_doc[f'__{attr}'] = doc[attr]
			elif type(doc[attr]) == dict:
				shadow_doc[attr] = self.format_doc_oper(doc=doc[attr])
			else:
				shadow_doc[attr] = doc[attr]
		return shadow_doc
